
import os
from typeguard import typechecked

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.oc.oc import OC, VMStatus
from benchmark_runner.main.environment_variables import environment_variables


class Virtctl(OC):
    """
    Interface to virtctl command
    """

    def __init__(self):
        super().__init__()

    @typechecked
    @logger_time_stamp
    def save_vm_log(self, vm_name: str, output_filename: str = '', namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        This method save vm log in log_path
        :param vm_name: vm name with uuid
        :param output_filename:
        """
        if not output_filename:
            output_filename = os.path.join(self._run_artifacts, vm_name)
        namespace = f'-n {namespace}' if namespace else ''
        self.run(cmd=f"virtctl console {namespace} {vm_name} > {output_filename}", background=True)

    @typechecked
    @logger_time_stamp
    def stop_vm_async(self, vm_name: str, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        The method stop vm
        @param vm_name:
        @param namespace:
        @return: command output
        """
        namespace = f'-n {namespace}' if namespace else ''
        return self.run(cmd=f'virtctl stop {vm_name} {namespace}')

    @typechecked
    @logger_time_stamp
    def stop_vm_sync(self, vm_name: str, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        The method stop vm and wait till stopped
        @param vm_name:
        @return: True or Error
        """
        self.stop_vm_async(vm_name=vm_name, namespace=namespace)
        return self.wait_for_vm_status(vm_name=vm_name, status=VMStatus.Stopped, namespace=namespace)

    @typechecked
    @logger_time_stamp
    def start_vm_async(self, vm_name: str, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        The method start vm
        @param vm_name:
        @param namespace:
        @return: command output
        """
        namespace = f'-n {namespace}' if namespace else ''
        return self.run(cmd=f'virtctl start {vm_name} {namespace}')

    @typechecked
    @logger_time_stamp
    def start_vm_sync(self, vm_name: str, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        The method start vm and wait till running
        @param vm_name:
        @return: True or Error
        """
        self.start_vm_async(vm_name=vm_name, namespace=namespace)
        return self.wait_for_vm_status(vm_name=vm_name, status=VMStatus.Running, namespace=namespace)

    @typechecked
    @logger_time_stamp
    def expose_vm(self, vm_name: str, namespace: str = environment_variables.environment_variables_dict['namespace']):
        """
        The method expose vm
        @param vm_name:
        """
        namespace = f'-n {namespace}' if namespace else ''
        self.run(cmd=f'virtctl expose vm {vm_name} --name {vm_name} {namespace} --port 27022 --target-port 22 --type NodePort')




import os

from benchmark_runner.common.oc.oc import OC
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.ocp_resources.create_ocp_resource_operations import CreateOCPResourceOperations


class CreateCustom(CreateOCPResourceOperations):
    """
    This class is created Custom resource
    """
    def __init__(self, oc: OC, path: str, resource_list: list):
        super().__init__(oc)
        self.__oc = oc
        self.__path = path
        self.__resource_list = resource_list

    @logger_time_stamp
    def create_custom(self):
        """
        This method create custom resource
        :return:
        """
        for resource in self.__resource_list:
            logger.info(f'run {resource}')
            if resource.endswith('.sh'):
                self.__oc.run(cmd=f'chmod +x {os.path.join(self.__path, resource)}; {self.__path}/./{resource}')
            else:
                self.__oc.create_async(yaml=os.path.join(self.__path, resource))




class BenchmarkRunnerError(Exception):
    """ Base class for all benchmark runner error classes.
        All exceptions raised by the benchmark runner library should inherit from this class. """
    pass


class CNVNotInstalled(BenchmarkRunnerError):
    """
    This class raises an error that CNV operator is not installed
    """
    def __init__(self, workload):
        self.message = f"{workload} requires CNV to be installed"
        super(CNVNotInstalled, self).__init__(self.message)


class KataNotInstalled(BenchmarkRunnerError):
    """
    This class raises an error that Kata operator is not installed
    """
    def __init__(self, workload):
        self.message = f"{workload} requires Kata to be installed"
        super(KataNotInstalled, self).__init__(self.message)


class ODFNotInstalled(BenchmarkRunnerError):
    """
    This class raises an error that ODF operator is not installed
    """
    def __init__(self, workload):
        self.message = f"{workload} requires ODF to be installed, set 'ODF_PVC' to False to run with Ephemeral"
        super(ODFNotInstalled, self).__init__(self.message)


class MissingScaleNodes(BenchmarkRunnerError):
    """
    This class raises an error for missing scale nodes
    """
    def __init__(self):
        self.message = "Missing scale nodes"
        super(MissingScaleNodes, self).__init__(self.message)


class MissingRedis(BenchmarkRunnerError):
    """
    This class raises an error for missing redis for scale synchronization
    """
    def __init__(self):
        self.message = "Missing redis"
        super(MissingRedis, self).__init__(self.message)


class MissingVMs(BenchmarkRunnerError):
    """
    This class raises an error for missing VMs
    """
    def __init__(self):
        self.message = "Missing running VMs"
        super(MissingVMs, self).__init__(self.message)




from typeguard import typechecked

from github import Github
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp
from benchmark_runner.main.environment_variables import environment_variables


class GitHubOperations:
    """
    This class responsible for GitHub operations
    """

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        github_token = self.__environment_variables_dict.get('github_token', '')
        self.g = Github(github_token)
        self.repo = self.g.get_repo(self.__environment_variables_dict.get('github_repository_short', ''))

    @typechecked
    def create_secret(self, secret_name: str, unencrypted_value: str):
        """
        This method create a new secret
        :return:
        """
        self.repo.create_secret(secret_name=secret_name, unencrypted_value=unencrypted_value)

    @typechecked
    def delete_secret(self, secret_name: str):
        """
        This method delete exist secret
        :return:
        """
        self.repo.delete_secret(secret_name=secret_name)



import time

from typeguard import typechecked
from azure.common.credentials import ServicePrincipalCredentials
from azure.mgmt.compute import ComputeManagementClient
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.oc.oc import OC

logger.setLevel('INFO')


class AzureOperations:
    """
    This class is responsible to all the operations against Azure vms
    """
    SLEEP_TIME = 300

    def __init__(self, azure_clientid: str, azure_secret: str, azure_tenantid: str, azure_subscriptionid: str, azure_resource_group_name: str, kubeadmin_password: str):
        self.__azure_clientid = azure_clientid
        self.__azure_secret = azure_secret
        self.__azure_tenantid = azure_tenantid
        self.__azure_subscriptionid = azure_subscriptionid
        self.__azure_resource_group_name = azure_resource_group_name
        self.__credentials = ServicePrincipalCredentials(client_id=self.__azure_clientid, secret=self.__azure_secret,
                                                         tenant=self.__azure_tenantid)
        self.__compute_client = ComputeManagementClient(self.__credentials, self.__azure_subscriptionid)
        self.__kubeadmin_password = kubeadmin_password

    @typechecked()
    @logger_time_stamp
    def restart_vm(self, vm_name: str = ''):
        """
        This method restarts azure vm
        :return:
        """
        vm_restart = self.__compute_client.virtual_machines.restart(self.__azure_resource_group_name, vm_name)
        logger.info('Waiting till VM restart')
        vm_restart.wait()
        vm_instance = self.__compute_client.virtual_machines.get(self.__azure_resource_group_name, vm_name)
        return vm_instance

    @typechecked()
    @logger_time_stamp
    def start_vm(self, vm_name: str = ''):
        """
        This method starts azure vm
        :return:
        """
        vm_start = self.__compute_client.virtual_machines.start(self.__azure_resource_group_name, vm_name)
        logger.info('Waiting till VM is running')
        vm_start.wait()
        vm_instance = self.__compute_client.virtual_machines.get(self.__azure_resource_group_name, vm_name)
        return vm_instance

    @typechecked()
    @logger_time_stamp
    def stop_vm(self, vm_name: str = ''):
        """
        This method stops azure vm
        :return:
        """
        vm_stop = self.__compute_client.virtual_machines.deallocate(self.__azure_resource_group_name, vm_name)
        logger.info('Waiting till VM Stopped (deallocated)')
        vm_stop.wait()
        vm_instance = self.__compute_client.virtual_machines.get(self.__azure_resource_group_name, vm_name)
        return vm_instance

    @typechecked()
    @logger_time_stamp
    def get_vm_status(self, vm_name: str = ''):
        """
        This method return vm status
        :return:
        """
        vm_instance = self.__compute_client.virtual_machines.get(self.__azure_resource_group_name, vm_name)
        return vm_instance

    @logger_time_stamp
    def oc_login(self):
        """
        This method login to the cluster with new credentials
        :return:
        """
        oc = OC(kubeadmin_password=self.__kubeadmin_password)
        oc.login()
        return oc

    def verify_sno_cluster_is_up(self):
        """
        This method verifies sno cluster is up and running
        @return:
        """
        master_nodes = self.oc_login().get_master_nodes()
        if len(master_nodes.split()) == 1:
            return True
        else:
            return False

    def start_cluster_with_verification(self, vm_name: str, max_retry=3):
        """
        This method starts cluster with verification
        @param vm_name:
        @param max_retry: 3
        @return:
        """
        for attempt in range(1, max_retry + 1):
            logger.info(f"Attempt {attempt} to start the cluster...")
            try:
                logger.info('Start VM')
                self.start_vm(vm_name=vm_name)
                logger.info(f"Wait {self.SLEEP_TIME} seconds till cluster is up and running")
                time.sleep(self.SLEEP_TIME)
                # Check if the cluster is up
                if self.verify_sno_cluster_is_up():
                    logger.info("Cluster is up and running!")
                    return True
                else:
                    logger.info("Cluster is not up yet.")
            except Exception as e:
                logger.error(f"Error occurred during attempt {attempt} to start the cluster: {str(e)}")
                logger.info('Stop Cluster VM')
                self.stop_vm(vm_name=vm_name)

        logger.error(f"Failed to start the cluster after {max_retry} attempts.")
        return False




import os
import pandas as pd
from typeguard import typechecked

# display bokeh
from bokeh.plotting import figure, show
from bokeh.io import output_notebook, output_file
from bokeh.models import HoverTool, Label, ColumnDataSource, LabelSet, Circle, NumeralTickFormatter


from benchmark_runner.jupyterlab.templates.logs_operations.logs_operations import LogsOperations
from benchmark_runner.jupyterlab.templates.analyze_workloads.visualize_workload_operations import VisualizeWorkloadOperations


class AnalyzeUperfLogs(VisualizeWorkloadOperations):
    """
    This class analyzes Uperf logs
    """

    def __init__(self, s3_logs_url: str):
        super().__init__()
        self.__workload = 'uperf'
        self.s3_logs_url = s3_logs_url
        self.__logs_operations = LogsOperations(s3_logs_url=self.s3_logs_url)

    def get_uperf_log_file(self):
        """
        This method returns uperf log file
        @return: uperf log file
        """
        workload_dir = self.__logs_operations.get_workload_dir()
        workload_files_names = [file for file in os.listdir(workload_dir) if
                                file.startswith(f'{self.__workload}-client')]
        workload_log_file = os.path.join(workload_dir, workload_files_names[0])
        return workload_log_file

    @typechecked
    def extract_uperf_result(self, workload_log_file: str):
        """
        This method extracts uperf result
        @return: uperf result
        """
        latency = 'uperf: 95%ile Latency(ms) :'
        bytes = 'Average byte : '
        iops = 'Average ops :'
        GB = 1000000000
        latency_list = []
        byte_list = []
        iops_list = []
        uperf_result = {}
        with open(workload_log_file, 'r') as file:
            for line in file:
                if latency in line:
                    latency_list.append(float(line.split()[-1]))
                if bytes in line:
                    byte_list.append(float(line.split()[-1]))
                if iops in line:
                    iops_list.append(float(line.split()[-1]))
        avg_latency = sum(latency_list) / len(latency_list)
        uperf_result['Latency'] = round(avg_latency, 2)
        avg_bytes = sum(byte_list) / len(byte_list)
        uperf_result['Throughput'] = round(avg_bytes / GB, 2)
        avg_iops = sum(iops_list) / len(iops_list)
        # uperf_result['IOPS'] = round(avg_iops,2)
        return uperf_result

    @typechecked
    def display_uperf_result(self, uperf_result: dict, workload_log_file: str):
        """
        This method displays uperf result
        @return:
        """
        # Configure Bokeh to display plots in the notebook
        output_notebook()

        # Create a DataFrame from the dictionary
        df = pd.DataFrame.from_dict(uperf_result, orient='index', columns=['Value'])

        # Round the values to 2 decimal places
        df['Value'] = df['Value'].round(2)

        # Convert the index to a list of strings
        df['Metric'] = df.index.tolist()

        # Create a Bokeh figure
        p = figure(x_range=df['Metric'].tolist(), width=VisualizeWorkloadOperations.WIDTH, title=workload_log_file.split('/')[-1], y_axis_label='Value')

        # Add vertical bars to the figure
        bars = p.vbar(x=df['Metric'], top=df['Value'], width=0.5)

        # Add labels to the bars
        labels = [f"{val:.2f}" for val in df['Value']]
        p.text(x=df['Metric'], y=df['Value'], text=labels, text_baseline="bottom", text_align="center",
               text_font_size='10pt', text_color='black')

        # Add hover tool with tooltips
        hover = HoverTool(tooltips=[('Metric', '@x'), ('Value', '@top{0.00}')], renderers=[bars])
        p.add_tools(hover)

        # Specify the output file
        output_file(os.path.join(f"{workload_log_file.split('.')[0]}.html"))

        # Display the graph
        show(p)

    def compare_results(self, result1: dict, result2: dict, legend_label1: str, legend_label2: str, block_size: str):
        """
        This method compare between 2 uperf results
        @param result1:
        @param result2:
        @param legend_label1:
        @param legend_label2:
        @param block_size: 64, 1024, 8192
        @return:
        """
        self.compare_run_results(result1=result1, result2=result2, legend_label1=legend_label1,
                                                         legend_label2=legend_label2,
                                                         title=f'Uperf block size {block_size}',
                                                         x_axis_label='Metric', y_axis_label='Rate')




# Custom exceptions must NOT expose data including hostnames, pathnames, IP addresses, etc. (security risk)
class SSHError(Exception):
    """ Base class for all SSH error classes.
        All exceptions raised by the benchmark runner library should inherit from this class. """
    pass


class SSHSubprocessError(SSHError):
    """
    This class exception is for Subprocess
    """
    def __init__(self):
        self.message = 'Subprocess ended with errors'
        super(SSHSubprocessError, self).__init__(self.message)



from typing import Optional


class ConnectionData:
    """ This class contains server connection details"""
    def __init__(self,
                 host_name: str,
                 user_name: str,
                 port: int,
                 timeout: int,
                 password: Optional[str] = None,
                 ssh_key: Optional[str] = None
                 ):
        self.__host_name = host_name
        self.__user_name = user_name
        self.__password = password
        self.__ssh_key = ssh_key
        self.__port = port
        self.__timeout = timeout

    @property
    def host_name(self):
        """
        host name
        :return:
        """
        return self.__host_name

    @property
    def user_name(self):
        """
        user name
        :return:
        """
        return self.__user_name

    @property
    def password(self):
        """
        password
        :return:
        """
        return self.__password

    @property
    def ssh_key(self):
        """
        ssh key
        :return:
        """
        return self.__ssh_key

    @property
    def port(self):
        """
        port of ssh, default 22
        :return:
        """
        return self.__port

    @property
    def timeout(self):
        """
        timeout in seconds
        :return:
        """
        return self.__timeout




import os
import re
import pandas as pd
from datetime import datetime, timedelta
import ipywidgets as widgets
from IPython.display import display, HTML
from typeguard import typechecked

from benchmark_runner.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from benchmark_runner.jupyterlab.templates.summary_report.summary_report_operations import SummaryReportOperations


class SummaryReportWidgets:
    """ This class is analyzing workload data and visualization """

    NETWORK_SPEED = os.getenv('network_speed')
    GRAFANA_URL = os.getenv('grafana_url')
    FETCH_OCP_VERSIONS_DAYS = int(os.getenv('fetch_ocp_versions_days'))
    FILTER_KIND = os.getenv('filter_kind')
    # take 1 day extra for end time
    DEFAULT_END_DATE = datetime.combine(datetime.now().date(), datetime.max.time())
    DEFAULT_START_DATE = datetime.combine(DEFAULT_END_DATE - timedelta(days=FETCH_OCP_VERSIONS_DAYS),
                                          datetime.min.time())
    STORAGE_TYPES = {'hammerdb': 'ODF', 'hammerdb_lso': 'LSO', 'uperf': 'Ephemeral', 'vdbench': 'ODF', 'vdbench_scale': 'ODF', 'bootstorm': 'Ephemeral', 'windows': 'ODF (virtio)'}
    compared_ocp_versions = None

    def __init__(self, elasticsearch: ElasticSearchOperations):
        self.elasticsearch = elasticsearch
        self.summary_report = SummaryReportOperations(self.elasticsearch)

    @typechecked()
    def set_fetch_ocp_version_days(self, num_fetch_days: int = FETCH_OCP_VERSIONS_DAYS):
        """
        This method updates the number of fetch days, also updates DEFAULT_START_DATE
        @param num_fetch_days:
        @return:
        """
        self.FETCH_OCP_VERSIONS_DAYS = num_fetch_days
        self.DEFAULT_START_DATE = datetime.combine(self.DEFAULT_END_DATE - timedelta(days=self.FETCH_OCP_VERSIONS_DAYS),
                                              datetime.min.time())

    @staticmethod
    def get_compared_ocp_versions():
        """
        This method returns changed global compared_ocp_versions
        @return:
        """
        try:
            return compared_ocp_versions[::-1]
        except NameError:
            # Handle the case when compared_ocp_versions is not defined
            return None

    def version_key(self, version):
        """
        This method splits the version string into parts and returns the numeric parts
        @param version:
        @return:
        """
        # Split the version string into parts: removing characters from versions like '4.X.0-rc.4' and '4.X.0-ec.4' to enable proper ordering
        parts = version.replace('-rc.', '-ec.').split('.')

        # Extract the numeric parts and convert them to integers
        numeric_parts = [int(re.search(r'\d+', part).group()) if re.search(r'\d+', part) else part for part in parts]

        return numeric_parts

    def get_ocp_distinct_list(self):
        """
        This method returns distinct list
        @return:
        """
        entries = self.elasticsearch.get_index_data_between_dates(index=f'ci-status',
                                                                  start_datetime=self.DEFAULT_START_DATE,
                                                                  end_datetime=self.DEFAULT_END_DATE)
        ocp_versions_list = [entry.get('_source').get('ocp_version') for entry in entries if
                             entry.get('_source').get('ocp_version') not in [None, 0]]
        distinct_ocp_versions = set(ocp_versions_list)
        ocp_versions_list = sorted(list(distinct_ocp_versions), key=self.version_key, reverse=True)
        return ocp_versions_list

    def get_two_last_major_versions(self):
        """
        This method returns last two major versions
        """
        versions = self.get_ocp_distinct_list()
        # Find the latest version
        latest_version = versions[0]

        # Calculate the version_before
        version_before = f"{latest_version.split('.')[0]}.{str(int(latest_version.split('.')[1]) - 1)}"

        # Filter versions that start with version_before
        matching_versions = [version for version in versions if version.startswith(version_before)]

        # Find the maximum minor version among the matching versions
        max_minor_version = max(
            [int(''.join(filter(str.isdigit, version.split('.')[2]))) for version in matching_versions],
            default=None
        )
        latest_version_before = f"{version_before}.{max_minor_version}" if max_minor_version is not None else version_before
        if self.get_compared_ocp_versions():
            return self.get_compared_ocp_versions()
        else:
            return (latest_version_before, latest_version)

    def choose_ocp_versions(self):
        """
        This method returns ocp versions widget
        @return:
        """

        # Create a prompt and a list widget
        ocp_version_widget = widgets.SelectMultiple(
            options=self.get_ocp_distinct_list(),
            disabled=False,
            layout=widgets.Layout(height='300px')  # Adjust the height here
        )

        # Initialize the compared_ocp_versions list
        initial_compared_ocp_versions = []
        # Set the default selected values in the widget
        ocp_version_widget.value = self.get_two_last_major_versions()
        compared_ocp_versions = self.get_two_last_major_versions()
        # Display the prompt and the widget
        display(ocp_version_widget)

        # Function to get selected values and update compared_ocp_versions
        def get_selected_versions(change):
            selected_versions = ocp_version_widget.value
            global compared_ocp_versions
            compared_ocp_versions = selected_versions

        # Attach the function to the widget's change event
        ocp_version_widget.observe(get_selected_versions, names='value')

    def get_workload_filtered_data(self, workload: str, filter_field: dict = None):
        """
        This method gets workload filtered data
        Located in this class for get_compared_ocp_versions
        @return:
        """
        entries = self.summary_report.get_workload_data(index=f'{workload}-results', start_datetime=self.DEFAULT_START_DATE,
                                                             end_datetime=self.DEFAULT_END_DATE)
        filter_df = self.summary_report.filter_data(entries=entries,
                                                    compared_ocp_versions=self.get_two_last_major_versions(),
                                                    kind=self.FILTER_KIND,
                                                    filter_field=filter_field)
        return filter_df

    def calc_median_geometric_mean_df(self, workload: str, filter_df: pd.DataFrame):
        """
        This method calculates median geometric mean
        @param workload:
        @param filter_df:
        @return:
        """
        return self.summary_report.calc_median_geometric_mean_df(workload=workload, df=filter_df)

    @typechecked
    def get_filtered_df(self, workload: str):
        """
        This method returns filtered dataframe between different OpenShift versions
        @return:
        """
        # Get workload filtered data
        if workload == 'hammerdb':
            filtered_df = self.get_workload_filtered_data(workload=workload, filter_field={'storage_type': 'odf'})
        elif workload == 'hammerdb_lso':
            filtered_df = self.get_workload_filtered_data(workload='hammerdb', filter_field={'storage_type': 'lso'})
        elif workload == 'vdbench':
            filtered_df = self.get_workload_filtered_data(workload=workload, filter_field={'scale': None})
        elif workload == 'vdbench_scale':
            # 'vdbench' index
            filtered_df = self.get_workload_filtered_data(workload='vdbench', filter_field={'scale': 6})
        # uperf, bootstorm
        else:
            filtered_df = self.get_workload_filtered_data(workload=workload)
        return filtered_df

    @typechecked
    def analyze_workload(self, workload: str):
        """
        This method analyzes workload metrics between different OpenShift versions
        @return:
        """
        median_geometric_mean_df = self.calc_median_geometric_mean_df(workload, self.get_filtered_df(workload))
        median_geometric_mean_df.rename(columns={'result': self.get_two_last_major_versions()}, inplace=True)
        return median_geometric_mean_df

    def analyze_all_workload(self, workloads: list = ['hammerdb', 'hammerdb_lso', 'uperf', 'vdbench', 'vdbench_scale', 'bootstorm', 'windows']):
        """
        This method analyzes all the workloads
        @param workloads:
        @return:
        """
        result_df = None
        for workload in workloads:
            median_geometric_mean_df = self.analyze_workload(workload=workload)

            # Add column types before comparison result
            median_geometric_mean_df.insert(len(median_geometric_mean_df.columns) - 1, 'storage type',
                                            median_geometric_mean_df['workload'].map(self.STORAGE_TYPES))

            result_df = pd.concat([result_df, median_geometric_mean_df], ignore_index=True)
        # change workloads names
        result_df.loc[result_df['workload'] == 'hammerdb_lso', 'workload'] = 'hammerdb'
        result_df.loc[result_df['workload'] == 'windows', 'workload'] = 'bootstorm'

        return result_df

    @staticmethod
    def color_percentage_values(val):
        """
        This method defines a function for coloring percentage values
        @param val:
        @return:
        """
        if pd.notna(val) and isinstance(val, (int, float)):
            if val < -50:
                color = 'red'
            elif val < -10:
                color = 'orange'
            elif val < 0:
                color = 'black'
            else:
                color = 'green'

            return f'color: {color}; text-align: center;'
        else:
            return ''  # Handle NaN values or non-numeric values

    def display_df(self, df: pd.DataFrame):
        """
        This method displays a style df with a colors: red/green to positive/negative results
        @return:
        """
        # Convert the last column to numeric values (including handling non-numeric values)
        last_column_name = df.columns[-1]
        df[last_column_name] = pd.to_numeric(df[last_column_name], errors='coerce')

        # align all values to the left
        styled_df = df.style.set_table_styles([
            {'selector': 'th:not(:last-child)', 'props': [('text-align', 'left')]},
            {'selector': 'td:not(:last-child)', 'props': [('text-align', 'left')]}
        ])
        # add % to last column values
        styled_df = styled_df.format({last_column_name: "{}%"})
        # add color to positive/negative
        styled_df = styled_df.map(self.color_percentage_values)
        # html
        network_speed = f'<div><span style="color: black; font-weight: bold;"> Network Speed: </span><span style="color: blue;">{self.NETWORK_SPEED}</span></div>'
        grafana_url = f'<div><span style="color: black; font-weight: bold;">** For more details Grafana url:</span> <a href="{self.GRAFANA_URL}" style="color: blue;" target="_blank">{self.GRAFANA_URL}</a></div>'
        # Display the HTML
        display(HTML(styled_df._repr_html_()))
        display(HTML(network_speed))
        display(HTML(grafana_url))

    def get_comparison_details(self):
        """
        This method returns comparison details versions, dates, uuid
        @return:
        """
        details_df = self.summary_report.extract_comparison_details(self.get_filtered_df(workload='hammerdb'))

        # Set display options to show all values without truncation
        pd.set_option('display.max_colwidth', None)

        # Use HTML and CSS to style the DataFrame for left alignment
        style_html = """
        <style>
        th, td {
            text-align: left !important;
        }
        </style>
        """

        # Display the HTML-styled DataFrame
        display(HTML(style_html + details_df.to_html(index=False)))




from benchmark_runner.common.grafana.update_grafana_latest_value_mappings import UpdateGrafanaLatestValueMappings
from benchmark_runner.main.environment_variables import environment_variables

# Update perf main.libsonnet with last versions
environment_variables_dict = environment_variables.environment_variables_dict
update_grafana_mappings_value = UpdateGrafanaLatestValueMappings(main_libsonnet_path=environment_variables_dict.get('main_libsonnet_path', ''))
last_versions = update_grafana_mappings_value.get_last_elasticsearch_versions()
update_grafana_mappings_value.update_value_mappings_last_versions(last_versions=last_versions)
update_grafana_mappings_value.update_main_libsonnet()




class OCError(Exception):
    """ Base class for all OC error classes.
        All exceptions raised by the benchmark runner library should inherit from this class. """
    pass


class LoginFailed(OCError):
    """This exception return login failed error"""
    def __init__(self):
        self.message = f'Login error, check credentials'
        super(LoginFailed, self).__init__(self.message)


class PodNotCreateTimeout(OCError):
    """This exception return pod create timeout error"""
    def __init__(self, pod_name):
        self.message = f'Pod name: {pod_name} does not create'
        super(PodNotCreateTimeout, self).__init__(self.message)


class PodNotInitializedTimeout(OCError):
    """This exception return pod initialized timeout error"""
    def __init__(self, workload):
        self.message = f'Pod : {workload} does not ready'
        super(PodNotInitializedTimeout, self).__init__(self.message)


class PodNotReadyTimeout(OCError):
    """This exception return pod ready timeout error"""
    def __init__(self, workload):
        self.message = f'Pod : {workload} does not ready'
        super(PodNotReadyTimeout, self).__init__(self.message)


class PodNotCompletedTimeout(OCError):
    """This exception return pod completed timeout error"""
    def __init__(self, workload):
        self.message = f'Pod with label: {workload} does not completed'
        super(PodNotCompletedTimeout, self).__init__(self.message)


class PodTerminateTimeout(OCError):
    """This exception return pod terminate timeout error"""
    def __init__(self, pod_name):
        self.message = f'Pod name: {pod_name} does not terminate'
        super(PodTerminateTimeout, self).__init__(self.message)


class PodNameNotExist(OCError):
    """This exception return pod name does not exist"""
    def __init__(self, pod_name):
        self.message = f'Pod name: {pod_name} does not exist'
        super(PodNameNotExist, self).__init__(self.message)


class VMNotCreateTimeout(OCError):
    """This exception return pod create timeout error"""
    def __init__(self, pod_name):
        self.message = f'VM name: {pod_name} is not created'
        super(VMNotCreateTimeout, self).__init__(self.message)


class VMDeleteTimeout(OCError):
    """This exception return vm delete timeout error"""
    def __init__(self, vm_name):
        self.message = f'VM name: {vm_name} is not deleted'
        super(VMDeleteTimeout, self).__init__(self.message)


class VMNameNotExist(OCError):
    """This exception return pod name does not exist"""
    def __init__(self, vm_name):
        self.message = f'VM name: {vm_name} does not exist'
        super(VMNameNotExist, self).__init__(self.message)


class YAMLNotExist(OCError):
    """This exception return pod name does not exist"""
    def __init__(self, yaml):
        self.message = f'YAML file: {yaml} does not exist'
        super(YAMLNotExist, self).__init__(self.message)


class VMNotInitializedTimeout(OCError):
    """This exception return vm initialized timeout error"""
    def __init__(self, workload):
        self.message = f'VM {workload} does not ready'
        super(VMNotInitializedTimeout, self).__init__(self.message)


class VMNotReadyTimeout(OCError):
    """This exception return vm ready timeout error"""
    def __init__(self, workload):
        self.message = f'VM : {workload} does not ready'
        super(VMNotReadyTimeout, self).__init__(self.message)


class VMStateTimeout(OCError):
    """This exception indicates timeout for VM state """
    def __init__(self, vm_name, state):
        self.message = f'VM: {vm_name} does not reach to state: {state}'
        super(VMStateTimeout, self).__init__(self.message)


class VMNotCompletedTimeout(OCError):
    """This exception return vm completed error"""
    def __init__(self, workload):
        self.message = f'VM : {workload} does not completed'
        super(VMNotCompletedTimeout, self).__init__(self.message)


class ExecFailed(OCError):
    """exec command on pod failed"""
    def __init__(self, pod, command, reason):
        self.message = f'exec {command} on {pod} failed: {reason}'
        super(ExecFailed, self).__init__(self.message)


class PodFailed(OCError):
    """exec command on pod failed"""
    def __init__(self, pod):
        self.message = f'pod {pod} failed'
        super(PodFailed, self).__init__(self.message)


class DVStatusTimeout(OCError):
    """This exception return dv status timeout error"""
    def __init__(self, status):
        self.message = f'DV status {status} timeout'
        super(DVStatusTimeout, self).__init__(self.message)


class CSVNotCreateTimeout(OCError):
    """This exception return csv create timeout error"""
    def __init__(self, operator, namespace):
        self.message = f'{operator} CSV in namespace: {namespace} is not created'
        super(CSVNotCreateTimeout, self).__init__(self.message)


class UpgradeNotStartTimeout(OCError):
    """This exception return ocp upgrade timeout error"""
    def __init__(self, version):
        self.message = f"OCP upgrade to {version} didn't start"
        super(UpgradeNotStartTimeout, self).__init__(self.message)


class OperatorInstallationTimeout(OCError):
    """This exception return operator installation timeout error"""
    def __init__(self, operator, version, namespace):
        self.message = f"{operator} operator installation to: {version} in namespace: {namespace} didn't complete"
        super(OperatorInstallationTimeout, self).__init__(self.message)


class OperatorUpgradeTimeout(OCError):
    """This exception return operator upgrade timeout error"""
    def __init__(self, operator, version, namespace):
        self.message = f"{operator} operator upgrade to: {version} in namespace: {namespace} didn't complete"
        super(OperatorUpgradeTimeout, self).__init__(self.message)




import os

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.elasticsearch.elasticsearch_exceptions import ElasticSearchDataNotUploaded
from benchmark_runner.main.environment_variables import environment_variables
from benchmark_runner.benchmark_operator.benchmark_operator_workloads_operations import BenchmarkOperatorWorkloadsOperations


class HammerdbVM(BenchmarkOperatorWorkloadsOperations):
    """
    This class runs hammerdb vm workload
    """
    ES_FETCH_TIME = 30

    def __init__(self):
        """
        All inherit from BenchmarkOperatorWorkloadsOperations
        """
        super().__init__()
        self.__name = ''
        self.__workload_name = ''
        self.__database = ''
        self.__es_index = ''
        self.__kind = ''
        self.__status = ''
        self.__es_fetch_min_time = self.ES_FETCH_TIME if self._run_type == 'perf_ci' else None

    @logger_time_stamp
    def run(self):
        """
        This method runs hammerdb vm workload
        :return:
        """
        try:
            self._prometheus_metrics_operation.init_prometheus()
            self.__name = f"{self._workload.split('_')[0]}_{self._workload.split('_')[1]}"
            self.__database = self._workload.split('_')[2]
            if self._run_type == 'test_ci':
                self.__es_index = 'hammerdb-test-ci-results'
            else:
                self.__es_index = 'hammerdb-results'
            self.__workload_name = self.__name.replace('_', '-')
            environment_variables.environment_variables_dict['kind'] = 'vm'
            self._oc.create_vm_sync(yaml=os.path.join(f'{self._run_artifacts_path}', f'{self.__name}_{self.__database}.yaml'), vm_name=f'{self.__workload_name}-workload')
            # hammerdb workload and database
            self._oc.wait_for_vm_create(vm_name=f'{self.__workload_name}-workload')
            self._oc.wait_for_initialized(label='app=hammerdb_workload', workload=self.__workload_name)
            self._oc.wait_for_ready(label='app=hammerdb_workload', workload=self.__workload_name)
            # Create vm log should be direct after vm is ready
            vm_name = self._create_vm_log(labels=[self.__workload_name])
            self.__status = self._oc.wait_for_vm_completed(workload=self.__workload_name, vm_name=vm_name)
            self.__status = 'complete' if self.__status else 'failed'
            # prometheus queries
            self._prometheus_metrics_operation.finalize_prometheus()
            metric_results = self._prometheus_metrics_operation.run_prometheus_queries()
            prometheus_result = self._prometheus_metrics_operation.parse_prometheus_metrics(data=metric_results)
            # system metrics
            if environment_variables.environment_variables_dict['system_metrics']:
                self.system_metrics_collector(workload=self.__workload_name, es_fetch_min_time=self.__es_fetch_min_time)
            # save run artifacts logs of benchmark-controller-manager and system-metrics
            run_artifacts_url = self._create_run_artifacts(workload=self.__workload_name, pod=False)
            # verify that data upload to elastic search
            if self._es_host:
                ids = self._verify_elasticsearch_data_uploaded(index=self.__es_index, uuid=self._oc.get_long_uuid(workload=self.__workload_name), es_fetch_min_time=self.__es_fetch_min_time)
                # update metadata
                for id in ids:
                    self._update_elasticsearch_index(index=self.__es_index, id=id, kind=self._environment_variables_dict.get('kind', ''), status=self.__status, run_artifacts_url=run_artifacts_url, database=self.__database, prometheus_result=prometheus_result)
            self._oc.delete_vm_sync(
                yaml=os.path.join(f'{self._run_artifacts_path}', f'{self.__name}_{self.__database}.yaml'),
                vm_name=f'{self.__workload_name}-workload')
        except ElasticSearchDataNotUploaded as err:
            self.tear_down_vm_after_error(yaml=os.path.join(f'{self._run_artifacts_path}', f'{self.__name}_{self.__database}.yaml'),
                                          vm_name=f'{self.__workload_name}-workload')
            raise err
        except Exception as err:
            # save run artifacts logs
            if self._oc.pod_exists(pod_name='benchmark-controller-manager'):
                self._create_pod_log(label='benchmark-controller-manager')
            full_name = f'{self.__workload_name}-{self.__database}'
            run_artifacts_url = os.path.join(self._environment_variables_dict.get('run_artifacts_url', ''), f'{self._get_run_artifacts_hierarchy(workload_name=full_name, is_file=True)}-{self._time_stamp_format}.tar.gz')
            if self._es_host:
                ids = self._verify_elasticsearch_data_uploaded(index=self.__es_index, uuid=self._oc.get_long_uuid(workload=self.__workload_name), timeout=3)
                if ids:
                    for id in ids:
                        self._update_elasticsearch_index(index=self.__es_index, id=id, kind=self.__kind, status='failed', run_artifacts_url=run_artifacts_url, database=self.__database)
                else:
                    self._upload_to_elasticsearch(index=self.__es_index, kind=self.__kind, status='failed', run_artifacts_url=run_artifacts_url, database=self.__database, uuid=self._uuid)
                    self._verify_elasticsearch_data_uploaded(index=self.__es_index, uuid=self._uuid)
            # delete hammerdb
            self.tear_down_vm_after_error(yaml=os.path.join(f'{self._run_artifacts_path}', f'{self.__name}_{self.__database}.yaml'),
                                          vm_name=f'{self.__workload_name}-workload')
            raise err




import os
import logging
from googleapiclient.discovery import build
from google.oauth2.credentials import Credentials
from googleapiclient.http import MediaFileUpload
from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import InstalledAppFlow
from benchmark_runner.common.logger.logger_time_stamp import logger  # Added logger import

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Define the scope
SCOPES = ['https://www.googleapis.com/auth/drive']


class GoogleDriveOperations:
    """
    This class is for Google Drive operations and requires the following:

    1. credentials.json - Obtain from the "Google Cloud Console" under "APIs & Services" > "Credentials" > "Create Credentials" and select "OAuth 2.0 Client ID".
    2. token.json - This is generated automatically. If a browser is not available, copy the token from a machine with a browser.
    """
    def __init__(self):
        """
        Initializes GoogleDriveOperations with authentication.
        """
        self.creds = self.authenticate()
        self.service = build('drive', 'v3', credentials=self.creds)

    def authenticate(self):
        """
        Authenticates and returns credentials using token.json or interactive login.
        :return: Credentials object
        """
        creds = None
        if os.path.exists('token.json'):
            creds = Credentials.from_authorized_user_file('token.json', SCOPES)
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file('credentials.json', SCOPES)
                creds = flow.run_local_server(port=0)
            with open('token.json', 'w') as token_file:
                token_file.write(creds.to_json())
        return creds

    def list_files_in_folder(self, folder_id, level=0):
        """
        Recursively lists files and folders in a specified Google Drive folder.
        :param folder_id: The ID of the folder to list.
        :param level: Current recursion level (default is 0).
        :return: List of file and folder metadata dictionaries.
        """
        items = []
        page_token = None
        while True:
            results = self.service.files().list(
                q=f"'{folder_id}' in parents and trashed = false",
                includeItemsFromAllDrives=True,
                supportsAllDrives=True,
                pageSize=100,
                fields="nextPageToken, files(id, name, mimeType)",
                pageToken=page_token
            ).execute()

            items.extend(results.get('files', []))
            page_token = results.get('nextPageToken', None)
            if page_token is None:
                break

        for item in items:
            indent = '  ' * level
            if item['mimeType'] == 'application/vnd.google-apps.folder':
                logger.info(f"{indent}Folder: {item['name']} ({item['id']})")
                self.list_files_in_folder(item['id'], level + 1)
            else:
                logger.info(f"{indent}File: {item['name']} ({item['id']})")

        return items

    def list_folders_and_files_in_shared_drive(self, shared_drive_id):
        """
        Lists all folders and files within a specified Google Drive shared drive.
        :param shared_drive_id: The ID of the shared drive to list.
        """
        try:
            logger.info(f'Folders and Files in shared drive with ID {shared_drive_id}:')
            self.list_files_in_folder(shared_drive_id)
            logger.info("Folder and file listing successful!")
        except Exception as e:
            logger.error(f"An error occurred: {e}")

    def find_existing_file(self, file_name, shared_drive_id):
        """
        Finds an existing file in a specified Google Drive shared drive.
        :param file_name: The name of the file to find.
        :param shared_drive_id: The ID of the shared drive to search.
        :return: File metadata dictionary if found, otherwise None.
        """
        try:
            all_files = self.list_files_in_folder(shared_drive_id)
            for file in all_files:
                if file['name'] == file_name:
                    return file
            return None
        except Exception as e:
            logger.error(f"An error occurred while searching for the file: {e}")
            return None

    def upload_file(self, file_path, shared_drive_id):
        """
        Uploads a file to a specified Google Drive shared drive and override if exist
        :param file_path: The local path of the file to upload.
        :param shared_drive_id: The ID of the shared drive to upload the file to.
        """
        try:
            file_name = os.path.basename(file_path)
            existing_file = self.find_existing_file(file_name, shared_drive_id)
            media = MediaFileUpload(file_path)

            if existing_file:
                file_id = existing_file['id']
                file = self.service.files().update(
                    fileId=file_id,
                    media_body=media,
                    supportsAllDrives=True,
                    fields='id'
                ).execute()
                logger.info(f"File '{file_name}' updated successfully in shared drive with ID: {shared_drive_id}")
            else:
                file_metadata = {
                    'name': file_name,
                    'parents': [shared_drive_id]
                }
                file = self.service.files().create(
                    body=file_metadata,
                    media_body=media,
                    supportsAllDrives=True,
                    fields='id'
                ).execute()
                logger.info(f"File '{file_name}' uploaded successfully to shared drive with ID: {shared_drive_id}")

        except Exception as e:
            logger.error(f"An error occurred: {e}")




import os

# Progress bar
import requests
import ipywidgets as widgets
from tqdm.auto import tqdm

# Display progress bar
from IPython.display import display

# logging
import logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)
from typeguard import typechecked


class LogsOperations:
    """
    This class responsible for logs operations
    """
    CHUNK_SIZE = 8192

    def __init__(self, s3_logs_url: str):
        # s3 Logs url is mandatory
        if not s3_logs_url:
            raise ValueError("s3_logs_url is empty")
        self.s3_logs_url = s3_logs_url
        self.filename = self.s3_logs_url.split('/')[-1]
        self.logs_dir = os.path.join(os.getcwd(), 'logs')
        self.log_dir_filename = os.path.join(self.logs_dir, self.filename)

    def cleanup(self):
        """
        This method cleans up existing logs and Prometheus images
        @return:
        """
        # Delete Prometheus container
        os.system('podman rmi -f docker.io/prom/prometheus;')
        # delete logs dir if exist
        if os.path.exists(self.logs_dir):
            os.system(f"rm -rf {self.logs_dir}")

    @typechecked
    def download_s3_logs(self, username: str, password: str):
        """
        This method downloads s3 logs
        @param username:
        @param password:
        @return:
        """
        if not os.path.exists(self.logs_dir):
            os.mkdir(self.logs_dir)

        # create a session with the credentials
        session = requests.Session()
        session.auth = (username, password)

        # download with progress bar
        response = session.get(self.s3_logs_url, stream=True)
        size = int(response.headers.get('Content-Length', 0))

        progress = widgets.IntProgress(description='Downloading', min=0, max=size)
        display(progress)

        with open(os.path.join(self.logs_dir, self.filename), 'wb') as f:
            with tqdm.wrapattr(f, "write", total=size) as fileobj:
                for chunk in response.iter_content(chunk_size=self.CHUNK_SIZE):
                    if chunk:
                        fileobj.write(chunk)
                        progress.value += len(chunk)

        logger.info('Download complete!')

    def untar_and_chmod_logs(self):
        """
        This method untars and sets the chmod for logs directory, and returns the path to the logs directory
        @return:
        """
        logger.info(f'untar download file {self.log_dir_filename}')
        os.system(f"tar -xvf {self.log_dir_filename} -C {self.logs_dir}")

    def get_workload_dir(self):
        """
        This method returns workload directory
        @return:
        """
        workload__directory = [folder for folder in os.listdir(self.logs_dir) if folder.startswith(self.filename.replace('.tar.gz','')) and not '.tar.gz' in folder]
        return os.path.join(self.logs_dir, workload__directory[0])




import inspect
import importlib

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.benchmark_operator.benchmark_operator_workloads_operations import BenchmarkOperatorWorkloadsOperations


class BenchmarkOperatorWorkloads(BenchmarkOperatorWorkloadsOperations):
    """
    This class runs benchmark-operator workload operations
    """
    def __init__(self):
        super().__init__()

    @logger_time_stamp
    def run(self):
        """
        The method runs benchmark-operator workload
        :return:
        """

        # kata use pod module - replace kata to pod
        workload = self._workload.replace('kata', 'pod')
        # if hammerdb split the database
        workload = f"{workload.split('_')[0]}_{workload.split('_')[1]}"
        # extract workload module and class
        workload_module = importlib.import_module(f'benchmark_runner.benchmark_operator.{workload}')

        try:
            self.initialize_workload()
            success = True
            for cls in inspect.getmembers(workload_module, inspect.isclass):
                if workload.replace('_', '').lower() == cls[0].lower():
                    if cls[1]().run() == False:
                        success = False

            self.finalize_workload()

            return success
        except Exception as err:
            self.finalize_workload()
            raise err




import os
import subprocess

# Open grafana url
from IPython.display import HTML, display

# parse grafana date
import time
from datetime import datetime

# logging
import logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)
from typeguard import typechecked

# log operations
from benchmark_runner.jupyterlab.templates.logs_operations.logs_operations import LogsOperations


class AnalyzePrometheusLogs:
    """
    This class analyze prometheus logs
    """
    TIMEOUT = 30
    SLEEP = 3

    def __init__(self, s3_logs_url: str):
        self.logs_operations = LogsOperations(s3_logs_url=s3_logs_url)

    def untar_and_chmod_prometheus_logs(self):
        """
        This method untars and sets the chmod for prometheus logs directory, and returns the path to the prometheus logs directory
        @return: prometheus_logs path
        """
        self.logs_operations.untar_and_chmod_logs()
        promdb_file = [f for f in os.listdir(self.logs_operations.log_dir_filename.split('.')[0]) if f.startswith('promdb')]
        logger.info(
            f"untar prometheus file: {os.path.join(self.logs_operations.logs_dir, self.logs_operations.filename.split('.')[0], promdb_file[0])}")
        os.system(
            f"tar -xvf {os.path.join(self.logs_operations.logs_dir, self.logs_operations.filename.split('.')[0], promdb_file[0])} -C {os.path.join(self.logs_operations.logs_dir, self.logs_operations.filename.split('.')[0])}")
        promdb_file = [f for f in os.listdir(self.logs_operations.log_dir_filename.split('.')[0]) if
                       f.startswith('promdb') and not f.endswith('tar')]
        promdb_dir_path = f"{os.path.join(self.logs_operations.logs_dir, self.logs_operations.filename.split('.')[0], promdb_file[0])}"

        logger.info(f'chmod {promdb_dir_path}')
        os.system(f"chmod -R g-s,a+rw {promdb_dir_path}")
        return promdb_dir_path

    @staticmethod
    @typechecked
    def run_container(image_name: str, command: str):
        """
        This method runs the container and waits until it finishes running
        @param image_name:
        @param command:
        @return:
        """
        cmd = f"podman run --name {image_name} {command}"
        logger.info(f"Running container image {image_name}")
        process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        current_wait_time = 0
        while current_wait_time <= AnalyzePrometheusLogs.TIMEOUT:
            output = process.stdout.readline()
            if output == b'' and process.poll() is not None:
                time.sleep(AnalyzePrometheusLogs.SLEEP)
                break
            if output:
                logger.info(output.strip())
            time.sleep(AnalyzePrometheusLogs.SLEEP)
            current_wait_time += AnalyzePrometheusLogs.SLEEP

        return_code = process.poll()
        logger.info(f"Container exited with return code {return_code}")
        return return_code

    @staticmethod
    @typechecked
    def open_grafana_dashboard(promdb_dir_path: str, grafana_dashboard_url: str):
        """
        This method opens the Grafana dashboard that is mounted to promdb
        @param promdb_dir_path:
        @param grafana_dashboard_url:
        @return:
        """
        start_time = promdb_dir_path.split('/')[-1].split('+')[:2][0].replace('prom', '').replace('db_', '')
        end_time = promdb_dir_path.split('/')[-1].split('+')[:2][1].replace('prom', '').replace('0000_', '')
        grafana_start_time = datetime.strptime(start_time, "%Y_%m_%dT%H_%M_%S")
        grafana_start_time = int(time.mktime(grafana_start_time.timetuple()) * 1000)
        grafana_end_time = datetime.strptime(end_time, "%Y_%m_%dT%H_%M_%S")
        grafana_end_time = int(time.mktime(grafana_end_time.timetuple()) * 1000)
        grafana_url = f"{grafana_dashboard_url}?orgId=1&from={grafana_start_time}&to={grafana_end_time}"
        logger.info(f"Grafana direct link:: {grafana_url}")
        js_code = f"window.open('{grafana_url}')"
        html_code = f"<script>{js_code}</script>"
        display(HTML(html_code))




import logging
import pandas as pd
from functools import reduce
from datetime import datetime
from typeguard import typechecked
from benchmark_runner.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations

# Configure the logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class SummaryReportOperations:
    """ This class is filtering data, aggregate data per workload and calculate geometric mean per run"""
    GB = 1_000_000_000
    MILLISECONDS = 1000
    HAMMERDB_METRIC = 'TPM'
    UPERF_THROUGHPUT_METRIC = 'Throughput'
    UPERF_LATENCY_METRIC = 'Latency'
    VDBENCH_LATENCY_METRIC = 'Latency'
    VDBENCH_IOPS_METRIC = 'Iops'
    BOOTSTORM_FEDORA = 'fedora37'
    BOOTSTORM_WINDOWS = ['windows10', 'windows11', 'windows_server_2019', 'windows_server_2022']
    BOOTSTORM_FEDORA_METRIC = '240 VMs run time'
    BOOTSTORM_WINDOWS_METRIC = '111 VMs run time'

    def __init__(self, elasticsearch: ElasticSearchOperations):
        self.elasticsearch = elasticsearch
        self.geometric_mean_data = None
        self.geometric_mean_data_throughput = None
        self.geometric_mean_data_iops = None
        self.geometric_mean_data_latency = None
        self.bootstorm_df = None
        self.median_indices = None
        self.geometric_mean_df = None

    @typechecked
    def get_workload_data(self, start_datetime: datetime, end_datetime: datetime, index: str ='ci-status'):
        """
        @return: fetching workload data from elasticsearch
        """

        return self.elasticsearch.get_index_data_between_dates(index=index, start_datetime=start_datetime, end_datetime=end_datetime)

    @typechecked
    def filter_data(self, entries: list, compared_ocp_versions: tuple, kind: str, filter_field: dict = None):
        """
        This method filters elastic search data by filter fields
        @param entries:
        @param compared_ocp_versions:
        @param kind:
        @param filter_field:
        @return:
        """
        filter_data = [
            entry.get('_source') for entry in entries
            if (
                    entry.get('_source').get('ocp_version') in compared_ocp_versions and
                    entry.get('_source').get('kind') == kind and
                    (filter_field is None or entry.get('_source').get(
                        list(filter_field.keys())[0]) == list(filter_field.values())[0])
            )
        ]
        return pd.DataFrame(filter_data)

    @staticmethod
    def __geometric_mean(values: list):
        """
        This method calculates geometric means for input values
        @param values:
        @return: values geometric means
        """
        product = reduce(lambda x, y: x * y, values)
        mean = pow(product, 1 / len(values))
        return round(mean, 3)

    @staticmethod
    def __get_median(values: dict):
        """
        This method returns the median result
        @param values: all median values
        @return: return median result
        """
        median_result = (values['geometric_mean'] - values['geometric_mean'].median()).abs().idxmin()
        return pd.Series({'median_result': median_result})

    def __update_geometric_mean_data(self, subset: pd.DataFrame, unique_uuid: str):
        """
        This method updates geometric mean data
        @param subset: aggregate data by uuid - each uuid represent a run
        @param unique_uuid:
        @return:
        """
        self.geometric_mean_data['uuid'].append(unique_uuid)
        self.geometric_mean_data['ocp_version'].append(subset['ocp_version'].iloc[0])
        self.geometric_mean_data['cnv_version'].append(subset['cnv_version'].iloc[0])
        self.geometric_mean_data['odf_version'].append(subset['odf_version'].iloc[0])

    @typechecked()
    def update_hammerdb_data(self, subset: pd.DataFrame, unique_uuid: str):
        """
        This method updates hammerdb data
        @param subset: aggregate data by uuid - each uuid represent a run
        @param unique_uuid:
        @return:
        """
        logger.info(subset[['workload', 'db_type', 'ocp_version', 'current_worker', 'tpm']])
        geometric_mean_result = self.__geometric_mean(subset['tpm'].tolist())
        # Get uuid, ocp_version, cnv_version, odf_version, geometric_mean and append data to result DataFrame
        self.__update_geometric_mean_data(subset, unique_uuid)
        self.geometric_mean_data['geometric_mean'].append(geometric_mean_result)
        self.geometric_mean_data['db_type'].append(subset['db_type'].iloc[0])

    @typechecked()
    def update_uperf_data(self, subset: pd.DataFrame, unique_uuid: str):
        """
        This method updates uperf data
        @param subset: aggregate data by uuid - each uuid represent a run
        @param unique_uuid:
        @return:
        """
        logger.info(
            f"{subset[['workload', 'read_message_size', 'num_threads', 'test_type', 'ocp_version', 'norm_ltcy']]}")
        # Calculate latency and throughput
        average_latency_df = subset.groupby(
            ['uuid', 'ocp_version', 'cnv_version', 'odf_version', 'read_message_size', 'num_threads',
             'test_type'])['norm_ltcy'].mean().reset_index()
        average_latency_df = average_latency_df[average_latency_df['test_type'] == 'rr']
        average_throughput_df = subset.groupby(
            ['uuid', 'ocp_version', 'cnv_version', 'odf_version', 'read_message_size', 'num_threads',
             'test_type'])['norm_byte'].mean().reset_index()
        average_throughput_df = average_throughput_df[average_throughput_df['test_type'] == 'stream']
        average_throughput_df['norm_byte'] = (average_throughput_df['norm_byte'] * 8) / self.GB
        # logger.info(f"Displaying average_latency_df:\n{average_latency_df}")
        # logger.info(f"Displaying average_throughput_df:\n{average_throughput_df}")
        # Calculate geometric mean
        geo_mean_latency = self.__geometric_mean(average_latency_df['norm_ltcy'])
        geo_mean_throughput = self.__geometric_mean(average_throughput_df['norm_byte'])
        self.__update_geometric_mean_data(subset, unique_uuid)
        self.geometric_mean_data_throughput.update(self.geometric_mean_data)
        self.geometric_mean_data_throughput['geometric_mean'].append(geo_mean_throughput)
        self.geometric_mean_data_latency.update(self.geometric_mean_data)
        self.geometric_mean_data_latency['geometric_mean'].append(geo_mean_latency)

    @typechecked()
    def update_vdbench_data(self, subset: pd.DataFrame, unique_uuid: str):
        """
        This method updates vdbench data
        @param subset: aggregate data by uuid - each uuid represent a run
        @param unique_uuid:
        @return:
        """
        # Calculate iops and latency
        average_latency_df = \
            subset.groupby(['uuid', 'ocp_version', 'cnv_version', 'odf_version', 'Run', 'Threads'])[
                'Resp'].mean().reset_index()
        average_iops_df = \
            subset.groupby(['uuid', 'ocp_version', 'cnv_version', 'odf_version', 'Run', 'Threads'])[
                'Rate'].mean().reset_index()

        geo_mean_latency = self.__geometric_mean(average_latency_df['Resp'])
        geo_mean_iops = self.__geometric_mean(average_iops_df['Rate'])

        logger.info('geo_mean_latency', geo_mean_latency)
        logger.info('geo_mean_iops', geo_mean_iops)
        self.__update_geometric_mean_data(subset, unique_uuid)
        self.geometric_mean_data_iops.update(self.geometric_mean_data)
        self.geometric_mean_data_iops['geometric_mean'].append(geo_mean_iops)
        self.geometric_mean_data_latency.update(self.geometric_mean_data)
        self.geometric_mean_data_latency['geometric_mean'].append(geo_mean_latency)

    @typechecked()
    def update_bootstorm_data(self, subset: pd.DataFrame, unique_uuid: str):
        """
        This method updates bootstorm data
        @param subset: aggregate data by uuid - each uuid represent a run
        @param unique_uuid:
        @return:
        """
        # Calculate bootstorm total run time
        self.bootstorm_df = subset.groupby(['vm_os_version', 'uuid', 'ocp_version', 'cnv_version', 'odf_version'])['total_run_time'].mean().reset_index()
        logger.info('geometric_mean_df', self.geometric_mean_df)
        self.__update_geometric_mean_data(subset, unique_uuid)
        self.geometric_mean_data['geometric_mean'].append(self.bootstorm_df['total_run_time'].iloc[0])
        self.geometric_mean_data['metric'].append(self.bootstorm_df['vm_os_version'].iloc[0])

    @typechecked()
    def prepare_workload_dataframe(self, workload: str, df: pd.DataFrame):
        """
        This method prepares workload dataframe
        @param workload:
        @param df
        @return:
        """
        # Create a DataFrame to store results
        # Initialize common structure
        common_structure = {'uuid': [], 'ocp_version': [], 'cnv_version': [], 'odf_version': []}

        if workload in ['hammerdb', 'hammerdb_lso']:
            self.geometric_mean_data = {'workload': workload, **common_structure, 'db_type': [], 'geometric_mean': []}
        elif workload == 'uperf':
            # Filter out values greater than 1000
            df = df[(df['norm_ltcy'] < 1000) & df['norm_ltcy'].notna()]
            self.geometric_mean_data = {**common_structure}
            self.geometric_mean_data_latency = {'workload': workload, **common_structure, 'metric': self.UPERF_LATENCY_METRIC,
                                                'geometric_mean': []}
            self.geometric_mean_data_throughput = {'workload': workload, **common_structure, 'metric': self.UPERF_THROUGHPUT_METRIC,
                                                   'geometric_mean': []}
        elif workload in ['vdbench', 'vdbench_scale']:
            df = df[df['Run'] != 'fillup']
            self.geometric_mean_data = {**common_structure}
            self.geometric_mean_data_iops = {'workload': workload, **common_structure, 'metric': self.VDBENCH_IOPS_METRIC,
                                             'geometric_mean': []}
            self.geometric_mean_data_latency = {'workload': workload, **common_structure, 'metric': self.VDBENCH_LATENCY_METRIC,
                                                'geometric_mean': []}
        elif workload in ['bootstorm', 'windows']:
            self.geometric_mean_data = {'workload': workload, **common_structure, 'metric': [], 'geometric_mean': []}

        return df

    @staticmethod
    def __calc_percentage(df: pd.DataFrame, complementary: bool = False):
        """
        This method calculates percentage difference
        @param df:
        @param complementary: calculates the complementary percentage change, meaning it measures the difference from 100%
        @return:
        """
        # complementary percentage change: e.g. latency
        if complementary:
            df['result'] = df.groupby('metric')['geometric_mean'].transform(lambda x: 1 - ((x - x.shift(1)) / x.shift(1)) * 100)
        else:
            df['result'] = df.groupby('metric')['geometric_mean'].transform(lambda x: ((x - x.shift(1)) / x.shift(1)) * 100)
        df = df.dropna(subset=['result'])
        df.loc[:, 'result'] = df['result'].round(2)
        return df[['workload', 'metric', 'result']]

    def __calc_workload_precentage_diff(self, workload: str, geometric_mean_data: dict, complementary: bool = False, metric: str = None):
        """
        This method calculates the percentage difference for workload dictionary and retrieving_operator_versions pre workload
        @workload:
        @geometric_mean_data: workload data including geometric_mean_data
        @complementary: how to calc the percentage
        @metric: which metric value to retrieve operator versions
        @return:
        """
        geometric_mean_df = pd.DataFrame(geometric_mean_data)
        self.median_indices = geometric_mean_df.groupby(['metric', 'ocp_version']).apply(self.__get_median)
        geometric_mean_df = geometric_mean_df.loc[self.median_indices['median_result']]
        return self.__calc_percentage(geometric_mean_df, complementary)

    def aggregate_hammerdb_dataframe(self, workload):
        """
        This method aggregates hammerdb dataframe and update metric names
        @return:
        """
        # rename db_type to metric to get same columns as other workloads
        self.geometric_mean_data['metric'] = self.geometric_mean_data.pop('db_type')
        self.geometric_mean_df = self.__calc_workload_precentage_diff(workload, self.geometric_mean_data, metric='pg')
        # updates metric names
        replace_dict = {'mariadb': f'{self.HAMMERDB_METRIC} (mariadb)', 'mssql': f'{self.HAMMERDB_METRIC} (mssql)', 'pg': f'{self.HAMMERDB_METRIC} (postgresql)'}
        self.geometric_mean_df['metric'] = self.geometric_mean_df['metric'].replace(replace_dict)

    def aggregate_uperf_dataframe(self, workload):
        """
        This method aggregates uperf dataframe and update metric names
        @return:
        """
        geometric_mean_df_throughput = self.__calc_workload_precentage_diff(workload, self.geometric_mean_data_throughput)
        geometric_mean_df_latency = self.__calc_workload_precentage_diff(workload, self.geometric_mean_data_latency, complementary=True)

        # Concatenate DataFrames along rows (assuming the keys are the same)
        self.geometric_mean_df = pd.concat([geometric_mean_df_throughput, geometric_mean_df_latency], ignore_index=True)

    def aggregate_vdbench_dataframe(self, workload):
        """
        This method aggregates vdbench dataframe and update metric names
        @return:
        """
        geometric_mean_df_iops = self.__calc_workload_precentage_diff(workload, self.geometric_mean_data_iops)
        geometric_mean_df_latency = self.__calc_workload_precentage_diff(workload, self.geometric_mean_data_latency, complementary=True)

        # Concatenate DataFrames along rows (assuming the keys are the same)
        self.geometric_mean_df = pd.concat([geometric_mean_df_iops, geometric_mean_df_latency], ignore_index=True)

    def aggregate_bootstorm_dataframe(self, workload: str):
        """
        This method aggregates bootstorm dataframe and update metric names
        @return:
        """
        self.geometric_mean_df = self.__calc_workload_precentage_diff(workload, self.geometric_mean_data, complementary=True, metric='windows_server_2019')
        # updates metric names
        if workload == 'windows':
            for os in self.BOOTSTORM_WINDOWS:
                self.geometric_mean_df['metric'] = self.geometric_mean_df['metric'].replace({os: f'{self.BOOTSTORM_WINDOWS_METRIC} ({os})'})
        else:
            self.geometric_mean_df['metric'] = self.geometric_mean_df['metric'].replace({f'{self.BOOTSTORM_FEDORA}': f'{self.BOOTSTORM_FEDORA_METRIC} ({self.BOOTSTORM_FEDORA})'})

    @typechecked()
    def aggregate_workload_dataframe(self, workload: str):
        """
        This method aggregates workload dataframe and updates metric values
        @param workload:
        @return:
        """
        if workload in ['hammerdb', 'hammerdb_lso']:
            self.aggregate_hammerdb_dataframe(workload)
        elif workload == 'uperf':
            self.aggregate_uperf_dataframe(workload)
        elif workload in ['vdbench', 'vdbench_scale']:
            self.aggregate_vdbench_dataframe(workload)
        elif workload in ['bootstorm', 'windows']:
            self.aggregate_bootstorm_dataframe(workload)

    @typechecked()
    def calc_median_geometric_mean_df(self, workload: str, df: pd.DataFrame):
        """
        This method calculates the geometric mean per workload from input dataframe
        @param workload:
        @param df:
        @return:
        """

        logger.info('Wait till fetch and analyzing data ...')
        # Convert lists to strings in the 'uuid' column
        df['uuid'] = df['uuid'].apply(
            lambda x: x[0] if isinstance(x, list) and x else x)

        df = self.prepare_workload_dataframe(workload, df)

        # Factorize the 'uuid' column
        labels, uniques = pd.factorize(df['uuid'])

        # Check if labels cover all rows
        if len(labels) != len(df):
            raise ValueError("Factorization did not cover all rows.")

        # Calculate geometric mean for each unique 'uuid'
        for label, unique_uuid in enumerate(uniques):
            # logger.info(f"For each UUID {unique_uuid}:")
            subset = df[labels == label]
            if workload in ['hammerdb', 'hammerdb_lso']:
                # Skip empty columns and skip NaN tpm values
                if not subset.empty and subset['tpm'].notna().all():
                    self.update_hammerdb_data(subset, unique_uuid)
            elif workload == 'uperf':
                if not subset.empty and subset['norm_ltcy'].notna().all() and subset['norm_byte'].notna().all():
                    self.update_uperf_data(subset, unique_uuid)
            elif workload in ['vdbench', 'vdbench_scale']:
                if not subset.empty and subset['Rate'].notna().all() and subset['Resp'].notna().all():
                    self.update_vdbench_data(subset, unique_uuid)
            elif workload in ['bootstorm', 'windows']:
                if not subset.empty and subset['total_run_time'].notna().all():
                    self.update_bootstorm_data(subset, unique_uuid)

        self.aggregate_workload_dataframe(workload)
        return self.geometric_mean_df

    @typechecked()
    def extract_comparison_details(self, df: pd.DataFrame):
        """
        This method returns comparison details ocp_version, odf_version, cnv_nightly_version, sample_dates and uuid
        @return:
        """
        df['timestamp'] = pd.to_datetime(df['timestamp'])
        df['date'] = df['timestamp'].dt.date
        aggregated_df = df.groupby('date').agg({
            'ocp_version': 'first',
            'odf_version': 'first',
            'cnv_version': 'first',
            'uuid': 'first'
        }).reset_index()

        aggregated_df = aggregated_df[['date', 'ocp_version', 'odf_version', 'cnv_version', 'uuid']]
        # Group by 'ocp_version' and aggregate every column with distinct values
        grouped_df = aggregated_df.groupby('ocp_version', as_index=False).agg(lambda x: list(set(x.explode())))

        grouped_df = grouped_df.rename(columns={'cnv_version': 'cnv_nightly_version'})
        grouped_df = grouped_df.rename(columns={'date': 'sample_dates'})

        return grouped_df[['ocp_version', 'odf_version', 'cnv_nightly_version', 'sample_dates', 'uuid']]



import os
import yaml
import time
import datetime
import openshift
from prometheus_api_client import PrometheusConnect
from urllib3.exceptions import InsecureRequestWarning
from urllib3 import disable_warnings
disable_warnings(InsecureRequestWarning)

from benchmark_runner.main.environment_variables import environment_variables
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger


class PrometheusMetricsOperation:
    """
    This class contains methods for running prometheus queries, parsing the results into dictionary nad uploading it into elasticsearch
    """

    def __init__(self):
        super().__init__()
        self._environment_variables_dict = environment_variables.environment_variables_dict
        self.__current_dir = os.path.dirname(os.path.abspath(__file__))
        self.__queries_file = os.path.join(self.__current_dir, 'metrics-default.yaml')
        self.__authorization = {'Authorization': self.__get_prometheus_token()}
        self.__prometheus = PrometheusConnect(url=self.__get_prometheus_default_url(), disable_ssl=True, headers=self.__authorization)
        self.__metrics_start_time = None
        self.__metrics_end_time = None
        self.__metric_results = {}
        self._prometheus_snap_interval = self._environment_variables_dict.get('prometheus_snap_interval', '')
        self._es_host = self._environment_variables_dict.get('elasticsearch', '')
        self._es_port = self._environment_variables_dict.get('elasticsearch_port', '')

    @staticmethod
    def __get_prometheus_default_url():
        """
        This method return prometheus default url
        :return:
        """
        try:
            with openshift.project('openshift-monitoring'):
                return f"https://{openshift.selector(['route/prometheus-k8s']).objects()[0].as_dict()['spec']['host']}"
        except Exception as err:
            raise f'Unable to retrieve prometheus-k8s route: {err}'

    def __get_prometheus_token(self):
        """
        This method return prometheus token
        :return:
        """
        try:
            return f'Bearer {self.__get_prometheus_token_by_version()}'
        except Exception as err:
            raise f'Unable to retrieve prometheus-k8s token: {err}'

    @staticmethod
    def __get_prometheus_token_by_version():
        """
        This method return prometheus token
        :return:
        """
        with openshift.project('openshift-monitoring'):
            openshift_version = openshift.get_server_version().split('-')[0].split('.')
            if int(openshift_version[0]) > 4 or (int(openshift_version[0]) == 4 and int(openshift_version[1]) > 10):
                return openshift.invoke('sa', ['new-token', '-n', 'openshift-monitoring', 'prometheus-k8s']).out().strip()
            else:
                return openshift.get_serviceaccount_auth_token('prometheus-k8s')

    @staticmethod
    def get_prometheus_timestamp():
        """
        This method return prometheus time stamp
        :return:
        """
        retries = 5
        while retries > 0:
            retries = retries - 1
            try:
                with openshift.project('openshift-monitoring'):
                    result = openshift.selector('pod/prometheus-k8s-0').object().execute(['date', '+%s.%N'],
                                                                                         container_name='prometheus')
                    return datetime.datetime.fromtimestamp(float(result.out()))
            except Exception as err:
                if retries <= 0:
                    raise f'Unable to retrieve date: {err}'
                else:
                    time.sleep(5)

    def yaml_to_dict(self):
        """
        This method convert yaml file to dictionary
        :return:
        """
        with open(self.__queries_file) as metrics_yaml:
            queries = metrics_yaml.read()
            return yaml.safe_load(queries)

    def init_prometheus(self):
        """
        This method init prometheus time
        :return:
        """
        self.__metrics_start_time = self.get_prometheus_timestamp()
        logger.info(f'prometheus metric start: {self.__metrics_start_time}')
        return self.__metrics_start_time

    def finalize_prometheus(self):
        """
        This method init prometheus time
        :return:
        """
        self.__metrics_end_time = self.get_prometheus_timestamp()
        logger.info(f'prometheus metric end: {self.__metrics_end_time}')
        return self.__metrics_end_time

    def run_prometheus_query(self, query: str, instant: bool = False):
        """
        This method run one prometheus query and return dictionary result
        :param instant:
        :param query:
        :return:
        """
        if instant:
            metric_result = (self.__prometheus.custom_query(query=query))
        else:
            metric_result = self.__prometheus.custom_query_range(query=query, start_time=self.__metrics_start_time, end_time=self.__metrics_end_time, step='30')
        self.__metric_results['metrics_result'] = metric_result
        return self.__metric_results

    def run_prometheus_queries(self, query_yaml_file: str = None):
        """
        This method run several prometheus queries and return list of dictionaries result
        :param query_yaml_file:
        :return:
        """
        if query_yaml_file and os.path.isfile(query_yaml_file):
            self.__queries_file = query_yaml_file
        metrics = self.yaml_to_dict()
        for metric in metrics:
            if 'query' not in metric:
                continue
            try:
                if 'instant' not in metric or metric['instant'] is not True:
                    metric_result = self.__prometheus.custom_query_range(metric['query'],
                                                                         start_time=self.__metrics_start_time,
                                                                         end_time=self.__metrics_end_time,
                                                                         step=self._prometheus_snap_interval)
                else:
                    metric_result = (self.__prometheus.custom_query(metric['query']))
                self.__metric_results[metric['metricName']] = metric_result
            except Exception as err:
                raise f"Query {metric['metricName']} ({metric['query']}) failed: {err}"
        return self.__metric_results

    #  @todo TBD: verify before uploading if it visualize correctly
    def upload_result_to_elastic(self):
        """
        This method upload result to ElasticSearch
        :return:
        """
        pass
        if self._es_host and self._es_port:
            for query, results in self.__metric_results.items():
                print(query, results)
        else:
            raise Exception('Missing ElasticSearch data')

    @staticmethod
    @logger_time_stamp
    def parse_prometheus_metrics(data: dict):
        """
        This method parses prometheus metrics and returns summary result
        based on queries: /benchmark_runner/common/prometheus/metrics-default.yaml
        @return:
        """
        result_dict = {}
        for query, data_list in data.items():
            if 'containerCPU-benchmark-runner' in query:
                suffix = 'CPU'
            elif 'containerMemory-benchmark-runner' in query:
                suffix = 'Memory'
            else:
                suffix = None
            total = 0
            max = 0
            for item in data_list:
                for val in item['values']:
                    if float(val[1]) > max:
                        max = round(float(val[1]), 3)
                total = total + max
                if not suffix:
                    result_dict[f'{query}'] = max
                else:
                    result_dict[f"{item['metric']['node']}_{suffix}"] = max
                    result_dict[f'total_{suffix}'] = total
        return result_dict




from functools import wraps

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.prometheus.prometheus_metrics_operations import PrometheusMetricsOperation


def prometheus_metrics(yaml_full_path: str = None):
    """
    This decorator method run prometheus metrics
    :param yaml_full_path: custom metric full path of yaml file [optional]
    :return:
    """
    def prometheus_metrics_internal(method):
        """
        This decorator method is for running prometheus queries and get prometheus metrics
        """
        @wraps(method)  # solve method help doc
        def method_wrapper(*args, **kwargs):
            prometheus_metrics_operation = PrometheusMetricsOperation()
            prometheus_metrics_operation.init_prometheus()
            try:
                result = method(*args, **kwargs)
                prometheus_metrics_operation.finalize_prometheus()
                # run several queries from custom yaml queries file or default yaml queries  file
                prometheus_metrics_operation.run_prometheus_queries(query_yaml_file=yaml_full_path)
                prometheus_metrics_operation.upload_result_to_elastic()
            except Exception as err:
                logger.error(f'error :{err}')
                raise err  # MethodError(method.__name__, err)

            return result
        return method_wrapper
    return prometheus_metrics_internal




class PrometheusSnapshotError(Exception):
    """
    Base class for all Prometheus snapshot exceptions
    """
    pass


class PrometheusSnapshotAlreadyStarted(PrometheusSnapshotError):
    """Snapshot already started"""
    def __init__(self):
        self.message = f'Attempt to start Prometheus snapshot when it was already started'
        super(PrometheusSnapshotAlreadyStarted, self).__init__(self.message)


class PrometheusSnapshotNotStarted(PrometheusSnapshotError):
    """Snapshot not started"""
    def __init__(self):
        self.message = f'Attempt to retrieve Prometheus snapshot when it was not started'
        super(PrometheusSnapshotNotStarted, self).__init__(self.message)


class PrometheusSnapshotAlreadyRetrieved(PrometheusSnapshotError):
    """Snapshot was already retrieved"""
    def __init__(self):
        self.message = f'Attempt to retrieve Prometheus snapshot when it was already retrieved'
        super(PrometheusSnapshotAlreadyRetrieved, self).__init__(self.message)




class LoggerError(Exception):
    """ Base class for all logger error classes.
        All exceptions raised by the benchmark runner library should inherit from this class. """
    pass


class MethodError(LoggerError):
    """
    This class is fot method error
    """
    def __init__(self, method_name, exception):
        self.message = f'method error: {method_name}, exception: {exception}'
        super(MethodError, self).__init__(self.message)




import ast  # change string list to list
from enum import Enum
from multiprocessing import set_start_method
import platform

from benchmark_runner.main.environment_variables import *
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.benchmark_operator.benchmark_operator_workloads import BenchmarkOperatorWorkloads
from benchmark_runner.workloads.workloads import Workloads
from benchmark_runner.main.environment_variables import environment_variables
from benchmark_runner.common.clouds.Azure.azure_operations import AzureOperations
from benchmark_runner.common.clouds.IBM.ibm_operations import IBMOperations
from benchmark_runner.common.clouds.BareMetal.bare_metal_operations import BareMetalOperations
from benchmark_runner.clusterbuster.clusterbuster_workloads import ClusterBusterWorkloads

# logger
log_level = os.environ.get('log_level', 'INFO').upper()
logger.setLevel(level=log_level)


SYSTEM_EXIT_BENCHMARK_FAILED = 1
SYSTEM_EXIT_UNKNOWN_EXECUTION_TYPE = 2

benchmark_operator_workload = None
benchmark_runner_workload = None
clusterbuster_workload = None

environment_variables_dict = environment_variables.environment_variables_dict
# environment variables data
provision_user = environment_variables_dict.get('provision_user', '')
workload = environment_variables_dict.get('workload', '')
azure_cluster_operation = environment_variables_dict.get('azure_cluster_operation', '')
ci_status = environment_variables_dict.get('ci_status', '')
install_ocp_version = environment_variables_dict.get('install_ocp_version', '')
upgrade_ocp_version = environment_variables_dict.get('upgrade_ocp_version', '')
odf_version = environment_variables_dict.get('odf_version', '')
lso_version = environment_variables_dict.get('lso_version', '')
cnv_version = environment_variables_dict.get('cnv_version', '')
install_ocp_resources = environment_variables_dict.get('install_ocp_resources', False)
run_type = environment_variables_dict.get('run_type', '')

is_benchmark_operator_workload = 'benchmark-operator' in (
environment_variables.get_workload_namespace(workload), environment_variables_dict.get("runner_type"))
is_benchmark_runner_workload = 'benchmark-runner' in (
environment_variables.get_workload_namespace(workload), environment_variables_dict.get("runner_type"))
is_clusterbuster_workload = 'clusterbuster' in (
environment_variables.get_workload_namespace(workload), environment_variables_dict.get("runner_type"))
# workload name validation
if workload and not ci_status:
    if workload not in environment_variables.workloads_list:
        logger.info(f'Enter valid workload name {environment_variables.workloads_list}')
        raise Exception(
            f'Not valid workload name: {workload} \n, choose one from the list: {environment_variables.workloads_list}')
    # run type validation
    if run_type not in environment_variables.run_types_list:
        logger.info(f'Enter valid run type {environment_variables.run_types_list}')
        raise Exception(
            f'Invalid run type: {run_type} \n, choose one from the list: {environment_variables.run_types_list}')
    if is_clusterbuster_workload:
        clusterbuster_workload = ClusterBusterWorkloads()
    elif is_benchmark_operator_workload:
        benchmark_operator_workload = BenchmarkOperatorWorkloads()
    elif is_benchmark_runner_workload:
        benchmark_runner_workload = Workloads()


# Define the Enum for cluster operations
class ClusterOperation(Enum):
    START = 'START'
    STOP = 'STOP'


@logger_time_stamp
def azure_cluster_operations(cluster_operation: str):
    """
    This method starts/stops Azure cluster according to input parameter azure_cluster_operation
    @param cluster_operation: START/ STOP
    @return:
    """
    azure_operation = AzureOperations(
                                      azure_clientid=environment_variables_dict.get('azure_clientid', ''),
                                      azure_secret=environment_variables_dict.get('azure_secret', ''),
                                      azure_tenantid=environment_variables_dict.get('azure_tenantid', ''),
                                      azure_subscriptionid=environment_variables_dict.get('azure_subscriptionid', ''),
                                      azure_resource_group_name=environment_variables_dict.get('azure_resource_group_name', ''),
                                      kubeadmin_password=environment_variables_dict.get('kubeadmin_password', '')
                                      )
    azure_vm_name = (environment_variables_dict.get('azure_vm_name', ''))
    if cluster_operation == ClusterOperation.START.value:
        logger.info('Start cluster with verification')
        azure_operation.start_cluster_with_verification(vm_name=azure_vm_name)
    elif cluster_operation == ClusterOperation.STOP.value:
        logger.info('Stop cluster')
        azure_operation.stop_vm(vm_name=azure_vm_name)
    else:
        raise Exception(f'Invalid cluster operation: {cluster_operation}')


@logger_time_stamp
def install_ocp_ibm_cloud(step: str):
    """
    This method runs IBM cloud OCP installation
    :return:
    """
    ibm_operations = IBMOperations(user=provision_user)
    ibm_operations.connect_to_provisioner()
    if step == 'run_ibm_ocp_installer':
        ibm_operations.update_ocp_version(file_name='ibmcloud.yml')
        ibm_operations.run_ocp_installer()
    # including update GitHub OCP credentials
    elif step == 'verify_ibm_install_complete':
        complete = ibm_operations.verify_install_complete()
        if complete:
            logger.info(f'OCP {install_ocp_version} installation complete successfully')
            logger.info(f'restart pod ci')
            ibm_operations.restart_pod_ci()
            logger.info(f'Update GitHub OCP credentials')
            ibm_operations.update_ocp_github_credentials()
        else:
            logger.info(f'OCP {install_ocp_version} installation failed')
    ibm_operations.disconnect_from_provisioner()


@logger_time_stamp
def install_ocp_bare_metal(step: str):
    """
    This method runs Bare Metal OCP installation
    :return:
    """
    bm_operations = BareMetalOperations(user=provision_user)
    bm_operations.connect_to_provisioner()
    if step == 'run_bare_metal_ocp_installer':
        bm_operations.update_ocp_version(file_name='all.yml')
        bm_operations.run_ocp_installer()
    elif step == 'verify_bare_metal_install_complete':
        complete = bm_operations.verify_install_complete()
        if complete:
            logger.info(f'OCP {install_ocp_version} installation complete successfully')
            logger.info(f'restart pod ci')
            bm_operations.restart_pod_ci()
        else:
            logger.info(f'OCP {install_ocp_version} installation failed')
    bm_operations.disconnect_from_provisioner()


@logger_time_stamp
def upgrade_ocp_bare_metal(step: str):
    """
    This method runs Bare Metal OCP upgrade
    :return:
    """
    bare_metal_operations = BareMetalOperations(user=provision_user)
    bare_metal_operations.connect_to_provisioner()
    oc = bare_metal_operations.oc_login()
    if step == 'run_bare_metal_ocp_upgrade':
        bare_metal_operations.run_ocp_upgrade(oc)
        # The LSO/ODF upgrade must be run manually after the OCP upgrade for the channel version; it won't upgrade automatically.
        bare_metal_operations.install_ocp_resources(resources=['lso'], upgrade_version=lso_version)
        bare_metal_operations.install_ocp_resources(resources=['odf'], upgrade_version=odf_version)
    elif step == 'verify_bare_metal_upgrade_complete':
        if bare_metal_operations.is_cluster_upgraded(oc, cnv_version=cnv_version, odf_version=odf_version, lso_version=lso_version):
            bare_metal_operations.verify_cluster_is_up(oc)
        else:
            error_message = f'OCP {upgrade_ocp_version} upgrade failed'
            logger.error(error_message)
            raise RuntimeError(error_message)
    bare_metal_operations.disconnect_from_provisioner()


@logger_time_stamp
def install_resources():
    """
    This method installs OpenShift resources
    :return:
    """
    bare_metal_operations = BareMetalOperations(user=provision_user)
    bare_metal_operations.connect_to_provisioner()
    install_resources_list = environment_variables_dict.get('install_resources_list', '')
    # convert str to list
    try:
        resources = ast.literal_eval(install_resources_list) if isinstance(install_resources_list, str) else install_resources_list
    except (ValueError, SyntaxError):
        resources = [install_resources_list]
    logger.info(f'Start Bare-Metal OpenShift resources installation')
    oc = bare_metal_operations.oc_login()
    bare_metal_operations.verify_cluster_is_up(oc)
    bare_metal_operations.install_ocp_resources(resources=resources)
    bare_metal_operations.disconnect_from_provisioner()
    logger.info(f'End Bare-Metal OpenShift resources installation')


@logger_time_stamp
def update_ci_status():
    """
    This method updates ci status
    :return:
    """
    ci_minutes_time = environment_variables_dict.get('ci_minutes_time', '')
    benchmark_runner_id = environment_variables_dict.get('benchmark_runner_id', '')
    benchmark_operator_id = environment_variables_dict.get('benchmark_operator_id', '')
    benchmark_wrapper_id = environment_variables_dict.get('benchmark_wrapper_id', '')
    benchmark_runner = Workloads()
    benchmark_runner.update_ci_status(status=ci_status, ci_minutes_time=int(ci_minutes_time),
                                      benchmark_runner_id=benchmark_runner_id,
                                      benchmark_operator_id=benchmark_operator_id,
                                      benchmark_wrapper_id=benchmark_wrapper_id)


@logger_time_stamp
def run_benchmark_operator_workload():
    """
    This method runs benchmark-operator workload
    :return:
    """
    # benchmark-operator node selector
    if environment_variables_dict.get('pin_node_benchmark_operator'):
        benchmark_operator_workload.update_node_selector(runner_path=environment_variables_dict.get('runner_path', ''),
                                                         yaml_path='benchmark-operator/config/manager/manager.yaml',
                                                         pin_node='pin_node_benchmark_operator')
    return benchmark_operator_workload.run()


@logger_time_stamp
def run_benchmark_runner_workload():
    """
    This method runs benchmark-runner workload
    :return:
    """
    # benchmark-runner node selector
    return benchmark_runner_workload.run()

@logger_time_stamp
def run_clusterbuster_workload():
    """
    This method runs clusterbuster workload
    :return:
    """
    # benchmark-runner node selector
    return clusterbuster_workload.run()


@logger_time_stamp
def main():
    """
    MAIN of benchmark-runner framework
    """
    success = True
    # START/ STOP Azure cluster
    if azure_cluster_operation:
        azure_cluster_operations(azure_cluster_operation)

    # install_ocp_version
    elif install_ocp_version:
        install_step = environment_variables_dict.get('install_step', '')
        logger.info(f'Starting installation step: {install_step}')
        if install_step == 'run_bare_metal_ocp_installer':
            install_ocp_bare_metal(step=install_step)
        elif install_step == 'run_ibm_ocp_installer':
            install_ocp_ibm_cloud(step=install_step)
        elif install_step == 'verify_bare_metal_install_complete':
            install_ocp_bare_metal(step=install_step)
        elif install_step == 'verify_ibm_install_complete':
            install_ocp_ibm_cloud(step=install_step)
        else:
            raise Exception(f'Invalid install step: {install_step}')
    # upgrade_ocp_version
    elif upgrade_ocp_version:
        upgrade_step = environment_variables_dict.get('upgrade_step', '')
        logger.info(f'Starting upgrading step: {upgrade_step}')
        if upgrade_step == 'run_bare_metal_ocp_upgrade':
            upgrade_ocp_bare_metal(step=upgrade_step)
        elif upgrade_step == 'verify_bare_metal_upgrade_complete':
            upgrade_ocp_bare_metal(step=upgrade_step)
        else:
            raise Exception(f'Invalid upgrade step: {upgrade_step}')
    # install_ocp_resource
    elif install_ocp_resources:
        install_resources()
    elif ci_status == 'pass' or ci_status == 'failed':
        update_ci_status()
    elif is_benchmark_operator_workload:
        success = run_benchmark_operator_workload()
    elif is_benchmark_runner_workload:
        success = run_benchmark_runner_workload()
    elif is_clusterbuster_workload:
        success = run_clusterbuster_workload()
    else:
        logger.error(f"empty workload, choose one from the list: {environment_variables.workloads_list}")
        raise SystemExit(SYSTEM_EXIT_UNKNOWN_EXECUTION_TYPE)

    if not success:
        logger.error("Benchmark failed.")
        raise SystemExit(SYSTEM_EXIT_BENCHMARK_FAILED)


if __name__ == '__main__':
    # spawn default in mac: https://docs.python.org/3/library/multiprocessing.html#:~:text=The%20default%20on%20Windows%20and%20macOS.
    if 'macos' in platform.platform(terse=True).lower():
        set_start_method('fork')
    main()




class BenchmarkOperatorError(Exception):
    """ Base class for all benchmark operator error classes.
        All exceptions raised by the benchmark runner library should inherit from this class. """
    pass


class CNVNotInstalled(BenchmarkOperatorError):
    """
    This class raises an error that CNV operator is not installed
    """
    def __init__(self, workload):
        self.message = f"{workload} requires CNV to be installed"
        super(CNVNotInstalled, self).__init__(self.message)


class KataNotInstalled(BenchmarkOperatorError):
    """
    This class raises an error that Kata operator is not installed
    """
    def __init__(self, workload):
        self.message = f"{workload} requires Kata to be installed"
        super(KataNotInstalled, self).__init__(self.message)


class ODFNotInstalled(BenchmarkOperatorError):
    """
    This class raises an error that ODF operator is not installed
    """
    def __init__(self, workload):
        self.message = f"{workload} requires ODF to be installed, set 'ODF_PVC' to False to run with Ephemeral"
        super(ODFNotInstalled, self).__init__(self.message)


class EmptyLSOPath(BenchmarkOperatorError):
    """
    This class raises an error that LSO path is empty
    """
    def __init__(self):
        self.message = "LSO path is empty"
        super(EmptyLSOPath, self).__init__(self.message)


class PrometheusSnapshotFailed(BenchmarkOperatorError):
    """
    This class raises an error when Prometheus snapshot failed
    """
    def __init__(self, err):
        self.message = f'Prometheus snapshot failed: {err}'
        super(PrometheusSnapshotFailed, self).__init__(self.message)




import csv
import os
import pandas as pd

# display df
from IPython.display import display
from typeguard import typechecked

# display bokeh
from bokeh.plotting import figure, show
from bokeh.io import output_notebook, output_file
from bokeh.models import HoverTool, Label, ColumnDataSource, LabelSet, Circle, NumeralTickFormatter

from benchmark_runner.jupyterlab.templates.logs_operations.logs_operations import LogsOperations
from benchmark_runner.jupyterlab.templates.analyze_workloads.visualize_workload_operations import VisualizeWorkloadOperations


class AnalyzeVdbenchLogs(VisualizeWorkloadOperations):
    """
    This class analyzes Vdbench logs
    """
    def __init__(self, s3_logs_url: str):
        super().__init__()
        self.__workload = 'vdbench'
        self.s3_logs_url = s3_logs_url
        self.__logs_operations = LogsOperations(s3_logs_url=self.s3_logs_url)

    @staticmethod
    @typechecked
    def is_csv_file(file_path: str):
        """
        This method checks if file is csv
        @param file_path:
        @return:
        """
        try:
            with open(file_path, 'r') as file:
                dialect = csv.Sniffer().sniff(file.read(1024))
                return dialect.delimiter == ',' or dialect.delimiter == ';'
        except csv.Error:
            return False

    def get_vdbench_csv_log_files(self):
        """
        This method returns vdbench csv log files path
        @return: vdbench csv log files path
        """
        workload_dir_name = self.__logs_operations.get_workload_dir()
        # Iterate over all files in the directory
        workload_files_path = []
        for file_name in os.listdir(workload_dir_name):
            file_path = os.path.join(workload_dir_name, file_name)
            if os.path.isfile(file_path) and self.is_csv_file(file_path) and self.__workload in file_name:
                workload_files_path.append(file_path)
        return workload_files_path

    def display_df(self):
        """
        This method display dataframe
        @return:
        """
        csv_files = self.get_vdbench_csv_log_files()
        for csv_file in csv_files:
            # Read the CSV file
            df = pd.read_csv(csv_file)
            df = df[df['Run'] != 'fillup']
            print(f"display file: {csv_file.split('/')[-1]}")
            display(df)

    def __display_horizontal(self, metric: str):
        """
        This method displays Vdbench metric horizontal
        @param metric: Latency/ IOPS
        @return:
        """
        metric_key, title = self.init_and_get_key_metric(metric)
        for csv_file in self.get_vdbench_csv_log_files():
            # Read the CSV file
            df = pd.read_csv(csv_file)
            df = df[df['Run'] != 'fillup']
            df = df.sort_values(metric_key)
            # Extract the desired columns from the DataFrame
            x = df['Run']
            y = df[metric_key]

            # Reverse the order of x values
            y = y[::-1]

            # Create a Bokeh figure
            p = figure(title=csv_file.split('/')[-1], y_range=x, x_axis_label='Latency (sec)',  width=VisualizeWorkloadOperations.WIDTH, height=VisualizeWorkloadOperations.HEIGHT)

            # Add a horizontal bar glyph to the figure with legend_label
            bars = p.hbar(y=x, right=y, height=0.5, legend_label=metric)

            # Add labels to the bars
            labels = [f"{val:,.2f}" for val in y]
            p.text(x=y, y=x, text=labels, text_baseline='middle', text_align='left', text_font_size='10pt')

            # Add hover tool with x-axis and y-axis values
            hover = HoverTool(tooltips=[('Run', '@y'), (metric, '@right{0,0.00}')], renderers=[bars])
            p.add_tools(hover)

            # Customize the legend
            p.legend.title = title
            p.legend.location = 'top_right'
            p.legend.orientation = 'vertical'
            p.legend.click_policy = 'hide'

            # Display the graph
            show(p)

    def display_iops_horizontal(self):
        """
        This method displays iops horizontal
        @return:
        """
        self.__display_horizontal(metric='IOPS')

    def display_latency_horizontal(self):
        """
        This method displays latency horizontal
        @return:
        """
        self.__display_horizontal(metric='Latency')

    def __display_vertical(self, metric: str):
        """
        This method displays Vdbench metric vertical
        @param metric: Latency/ IOPS
        @return:
        """
        metric_key, title = self.init_and_get_key_metric(metric)
        for csv_file in self.get_vdbench_csv_log_files():
            # Read the CSV file
            df = pd.read_csv(csv_file)
            df = df[df['Run'] != 'fillup']
            df = df.sort_values(metric_key)
            # Extract the desired columns from the DataFrame
            x = df['Run']
            y = df[metric_key]

            # Create a Bokeh figure
            p = figure(title=csv_file.split('/')[-1], x_range=x, y_axis_label=metric, width=VisualizeWorkloadOperations.WIDTH, height=VisualizeWorkloadOperations.HEIGHT)

            # Add vertical bars to the figure with legend_label
            bars = p.vbar(x=x, top=y, width=0.5, legend_label=metric)

            # Add labels to the bars
            labels = [f"{val:,.2f}" for val in y]
            p.text(x=x, y=y, text=labels, text_baseline='bottom', text_align='center', text_font_size='10pt')

            # Add hover tool with x-axis and y-axis values
            hover = HoverTool(tooltips=[('Run', '@x'), (metric, '@top{0,0.00}')], renderers=[bars])
            p.add_tools(hover)

            # Customize the y-axis tick formats
            p.yaxis.formatter = NumeralTickFormatter(format='0,0')

            # Customize the x-axis tick labels
            p.xaxis.major_label_orientation = 'vertical'
            p.xaxis.major_label_orientation = 45  # Rotate the labels by 45 degrees

            # Customize the legend
            p.legend.location = 'top_left'
            p.legend.orientation = 'vertical'
            p.legend.click_policy = 'hide'

            # Display the graph
            show(p)

    def display_iops_vertical(self):
        """
        This method displays iops vertical
        @return:
        """
        self.__display_vertical(metric='IOPS')

    def display_latency_vertical(self):
        """
        This method displays latency vertical
        @return:
        """
        self.__display_vertical(metric='Latency')

    @staticmethod
    def init_and_get_key_metric(metric: str):
        """
        This method init metric
        @param metric:
        @return:
        """
        if metric == 'Latency':
            metric_key = 'Resp'
            title = 'Latency - lower is better'
        elif metric == 'IOPS':
            metric_key = 'Rate'
            title = 'IOPS - higher is better'
        else:  # default Latency
            metric_key = 'Rate'
            title = 'IOPS - higher is better'
        # Configure Bokeh to display plots in the notebook
        output_notebook()
        return metric_key, title

    def __compare_results(self, csv_log_files1: str, csv_log_files2: str, legend_label1: str, legend_label2: str,
                          metric: str):
        """
        This method compares between artifacts run results
        @param csv_log_files1:
        @param csv_log_files2:
        @param legend_label1:
        @param legend_label2:
        @param metric:
        @return:
        """
        metric_key, title = self.init_and_get_key_metric(metric)
        # Read the CSV file into a DataFrame
        df1 = pd.read_csv(csv_log_files1[0])
        df2 = pd.read_csv(csv_log_files2[0])

        # Filter the DataFrame to include only the desired rows
        columns = ['oltp1', 'oltp2', 'oltphw', 'odss2', 'odss128', '4kb_read_16threads', '64kb_read_4threads',
                   '4kb_write_16threads', '64kb_write_4threads']
        df1 = df1[df1['Run'].isin(columns)]
        df2 = df2[df2['Run'].isin(columns)]

        # Set the index to the "Run" column
        df1.set_index('Run', inplace=True)
        df2.set_index('Run', inplace=True)

        # Convert the DataFrame to a dictionary Rate/Resp
        result1 = df1[metric_key].to_dict()
        result2 = df2[metric_key].to_dict()
        # Extract the keys and values from the dictionaries
        keys = list(result1.keys())
        values1 = list(result1.values())
        values2 = list(result2.values())

        # Calculate the x-coordinates for data1 and data2 bars
        x1 = [i - 0.2 for i in range(len(keys))]
        x2 = [i + 0.2 for i in range(len(keys))]

        # Shift the x-coordinates to the right switch between x1 to x2 to change the bar display order
        x1_shifted = [x + 0.5 for x in x2]
        x2_shifted = [x + 0.5 for x in x1]

        # Create a Bokeh figure
        p = figure(title=f'Vdbench {title}', x_range=keys, y_axis_label='Latency (sec)',  width=VisualizeWorkloadOperations.WIDTH, height=VisualizeWorkloadOperations.HEIGHT)

        # Adjust the y-axis range
        p.y_range.start = 0
        p.y_range.end = max(max(result1.values()), max(result2.values())) * 2
        p.yaxis.formatter = NumeralTickFormatter(format='0,')  # Format y-axis tick labels with commas

        # Add vertical bars for data1
        r1 = p.vbar(x=x2_shifted, top=values1, width=0.4, color='green', legend_label=legend_label1)

        # Add vertical bars for data2
        r2 = p.vbar(x=x1_shifted, top=values2, width=0.4, color='blue', legend_label=legend_label2)

        # Calculate the percentage difference between values1 and values2
        percentage_diff = [(v1 - v2) / v2 * 100 if v2 != 0 else 0 for v1, v2 in zip(values1, values2)]
        percentage_diff = [round(diff, 2) for diff in percentage_diff]  # Round to 2 decimal places

        # Customize the x-axis tick labels
        p.xaxis.major_label_orientation = 45  # Rotate the labels by 45 degrees
        p.xaxis.axis_label = 'Applications'  # Add x-axis title
        p.xaxis.major_label_text_font_size = '12pt'  # Increase the label font size

        # Add tooltips for data1 bars
        hover1 = HoverTool(renderers=[r1], tooltips=[(legend_label1, '@top{0,0.00}')])
        p.add_tools(hover1)

        # Add tooltips for data2 bars
        hover2 = HoverTool(renderers=[r2], tooltips=[(legend_label2, '@top{0,0.00}')])
        p.add_tools(hover2)

        # Add labels for result1 bars
        for i, value in enumerate(values2):
            label = Label(x=x1_shifted[i], y=value, text=f'{value:,.2f}', text_font_size='8pt', text_color='black',
                          text_baseline='bottom', text_align='center')
            p.add_layout(label)

        # Add labels for result2 bars
        for i, value in enumerate(values1):
            label = Label(x=x2_shifted[i], y=value, text=f'{value:,.2f}', text_font_size='8pt', text_color='black',
                          text_baseline='bottom', text_align='center')
            p.add_layout(label)

        # Calculate the maximum value from result1 and result2 dictionaries
        min_value = min(min(values1), min(values2))

        # Calculate the percentage difference between data1 and data2
        if metric == 'Latency':
            percentage_diff = [(v1 - v2) / v2 * 100 if v2 != 0 else 0 for v1, v2 in zip(values1, values2)]
        elif metric == 'IOPS':
            percentage_diff = [(v2 - v1) / v1 * 100 if v2 != 0 else 0 for v1, v2 in zip(values1, values2)]
        percentage_diff = [round(diff, 2) for diff in percentage_diff]  # Round to 2 decimal places

        # Calculate the x-coordinates for the dots
        x_middle = [(x1_shifted[i] + x2_shifted[i]) / 2 for i in range(len(keys))]

        # Calculate the y-coordinate for the dots
        y_middle = [max(values1[i], values2[i]) + min_value for i in range(len(keys))]

        # Add labels and dots for percentage difference
        for i, diff in enumerate(percentage_diff):
            x_pos = x_middle[i]
            y_pos = y_middle[i] + 0.1  # Adjust the y-coordinate to move the labels above the line
            color = "red" if diff < 0 else 'green'  # Set label and dot color based on the sign of the difference
            label = Label(x=x_pos, y=y_pos, text=f'{diff}%', text_font_size='14pt', text_color=color,
                          text_baseline='bottom', text_align='center')
            p.add_layout(label)

            # Adjust the y-coordinate to place the dots below the labels
            dot_y = y_pos - 0.1

            # Change dot color based on the sign of the difference
            dot_color = 'red' if diff < 0 else 'green'
            dot_source = ColumnDataSource(data=dict(x=[x_middle[i]], y=[dot_y], color=[dot_color]))
            p.circle(x='x', y='y', size=8, fill_color='color', line_color='color', source=dot_source)

        # Display the graph
        show(p)

    def compare_latency(self, csv_log_files1: str, csv_log_files2: str, legend_label1: str, legend_label2: str,
                        metric: str = 'Latency'):
        """
        This method compares latency results between logs
        @param csv_log_files1:
        @param csv_log_files2:
        @param legend_label1:
        @param legend_label2:
        @param metric:
        @return:
        """
        self.__compare_results(csv_log_files1, csv_log_files2, legend_label1, legend_label2, metric=metric)

    def compare_iops(self, csv_log_files1: str, csv_log_files2: str, legend_label1: str, legend_label2: str,
                     metric: str = 'IOPS'):
        """
        This method compares iops results between logs
        @param csv_log_files1:
        @param csv_log_files2:
        @param legend_label1:
        @param legend_label2:
        @param metric:
        @return:
        """
        self.__compare_results(csv_log_files1, csv_log_files2, legend_label1, legend_label2, metric=metric)




class S3OperationsError(Exception):
    """ Base class for all s3 operations error classes.
        All exceptions raised by the benchmark runner library should inherit from this class. """
    pass


class S3FileNotUploaded(S3OperationsError):
    """
    This class represents an error indicating that the S3 file was not uploaded
    """
    def __init__(self):
        self.message = "S3 file/object was not uploaded"
        super(S3FileNotUploaded, self).__init__(self.message)


class S3FileNotDownloaded(S3OperationsError):
    """
    This class represents an error indicating that the S3 file could not be downloaded
    """
    def __init__(self):
        self.message = "S3 file/object could not be downloaded"
        super(S3FileNotDownloaded, self).__init__(self.message)


class S3FileNotDeleted(S3OperationsError):
    """
    This class represents an error indicating that the S3 file/object was not deleted
    """
    def __init__(self):
        self.message = "S3 file/object is not deleted"
        super(S3FileNotDeleted, self).__init__(self.message)


class S3KeyNotCreated(S3OperationsError):
    """
    This class represents an error indicating that the S3 key/folder was not created.
    """
    def __init__(self):
        self.message = "S3 folder/key was not created"
        super(S3KeyNotCreated, self).__init__(self.message)


class S3FileNotExist(S3OperationsError):
    """
    This class represents an error indicating that the S3 file/object does not exist
    """
    def __init__(self):
        self.message = "S3 file/object does not exist"
        super(S3FileNotExist, self).__init__(self.message)


class S3FailedCreatePresingedURL(S3OperationsError):
    """
    This class is error that failed to create presigned url
    """
    def __init__(self):
        self.message = "failed to create presigned url"
        super(S3FailedCreatePresingedURL, self).__init__(self.message)




import os

from benchmark_runner.common.oc.oc import OC
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.ocp_resources.create_ocp_resource_operations import CreateOCPResourceOperations


class MigrateInfra(CreateOCPResourceOperations):
    """
    Migrate infrastructure to master nodes
    """
    def __init__(self, oc: OC, path: str, resource_list: list):
        super().__init__(oc)
        self.__oc = oc
        self.__path = path
        self.__resource_list = resource_list

    @logger_time_stamp
    def migrate_infra(self):
        """
        Actually migrate infrastructure
        :return: False if anything fails or if no master nodes are found.
        """
        for resource in self.__resource_list:
            logger.info(f'run {resource}')
            if resource == '01_cluster-monitoring-configmap-template.yaml':
                nodes = self.__oc.get_master_nodes().split()
                if nodes:
                    for node in nodes:
                        self.__oc.run(f'oc label node {node} node-role.kubernetes.io/infra=')
                    return self.__oc.create_async(yaml=os.path.join(self.__path, resource))
                else:
                    return False
        return False




import time
import ssl

from typeguard import typechecked
from elasticsearch import Elasticsearch
from elasticsearch.connection import create_ssl_context
from elasticsearch_dsl import Search
from datetime import datetime, timezone

from benchmark_runner.common.elasticsearch.elasticsearch_exceptions import ElasticSearchDataNotUploaded
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger


class ElasticSearchOperations:
    """
    This class contains elasticsearch operations
    """

    # sleep time between checks is 10 sec
    SLEEP_TIME = 10
    # ElasticSearch fetch data of last 15 minutes
    ES_FETCH_MIN_TIME = 15
    # max search results
    MAX_SEARCH_RESULTS = 1000
    MIN_SEARCH_RESULTS = 100

    def __init__(self, es_host: str, es_port: str, es_user: str = '', es_password: str = '', es_url_protocol: str = 'http', timeout: int = 2000):
        self.__es_host = es_host
        self.__es_port = es_port
        self.__es_user = es_user
        self.__es_password = es_password
        self.__es_url_protocol = es_url_protocol
        self.__timeout = timeout

        kwargs = {}
        if self.__es_url_protocol == 'https':
            ssl_context = create_ssl_context()
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE

            kwargs["ssl_context"] = ssl_context

        if self.__es_password:
            self.__es_url = f"{self.__es_url_protocol}://{self.__es_user}:{self.__es_password}@{self.__es_host}:{self.__es_port}"

            host_arg = self.__es_url
        else:
            host_arg = dict(host=self.__es_host)
            if self.__es_port:
                host_arg['port'] = self.__es_port

        self.__es = Elasticsearch([host_arg], **kwargs)

        self._hits = 0

    @property
    def hits(self):
        """This method is getter of hits"""
        return self._hits

    @hits.setter
    def hits(self, value):
        """This method is setter of hits"""
        try:
            index, uuid, workload = value
        except ValueError:
            raise ValueError("Pass an iterable with two items")
        self._hits = self.__elasticsearch_get_index_hits(index=index, uuid=uuid, workload=workload)

    def __elasticsearch_get_index_hits(self, index: str, uuid: str = '', fast_check: bool = False, id: bool = False, es_fetch_min_time:int = None):
        """
        This method searches for data per index in last 2 minutes and return the number of docs or zero
        :param index:
        :param es_fetch_min_time
        :param id: True to return the doc ids
        :param fast_check: return fast response
        :return:
        """
        """
        :return:
        """
        ids = []
        # https://github.com/elastic/elasticsearch-dsl-py/issues/49
        self.__es.indices.refresh(index=index)
        # timestamp name in Elasticsearch is different
        if es_fetch_min_time:
            search = Search(using=self.__es, index=index).filter('range', timestamp={
                'gte': f'now-{es_fetch_min_time}m', 'lt': 'now'})
        else:
            search = Search(using=self.__es, index=index).filter('range', timestamp={
                'gte': f'now-{self.ES_FETCH_MIN_TIME}m', 'lt': 'now'})
        # reduce the search result
        if fast_check:
            search = search[0:self.MIN_SEARCH_RESULTS]
        else:
            search = search[0:self.MAX_SEARCH_RESULTS]
        search_response = search.execute()
        if search_response.hits:
            if uuid:
                count_hits = 0
                for row in search_response:
                    if type(row['uuid']) == str:
                        # uperf return str
                        current_uuid = row['uuid']
                    else:
                        current_uuid = row['uuid'][0]
                    if current_uuid == uuid:
                        if fast_check:
                            return 1
                        ids.append(row.meta.id)
                        count_hits += 1
                if id:
                    return ids
                else:
                    return count_hits
            else:
                return len(search_response.hits)
        else:
            return 0

    @typechecked()
    @logger_time_stamp
    def verify_elasticsearch_data_uploaded(self, index: str, uuid: str = '', workload: str = '',
                                           fast_check: bool = False, timeout: int = None, es_fetch_min_time: int = None):
        """
        The method waits till data uploads to elasticsearch and waits if there is new data, search in last 15 minutes
        :param es_fetch_min_time:
        :param index:
        :param uuid: the current workload uuid
        :param workload: workload name only if there is a different timestamp parameter name in elasticsearch
        :param timeout:
        :param fast_check: return response on first doc

        :return:
        """
        current_wait_time = 0
        current_hits = 0
        self.__timeout = timeout if timeout else self.__timeout
        # waiting for any hits
        while self.__timeout <= 0 or current_wait_time <= self.__timeout:
            # waiting for new hits
            new_hits = self.__elasticsearch_get_index_hits(index=index, uuid=uuid, fast_check=fast_check, es_fetch_min_time=es_fetch_min_time)
            if current_hits < new_hits:
                logger.info(f'Data with index: {index} and uuid={uuid} was uploaded to ElasticSearch successfully')
                return self.__elasticsearch_get_index_hits(index=index, uuid=uuid, id=True, fast_check=fast_check, es_fetch_min_time=es_fetch_min_time)
            # sleep for x seconds
            time.sleep(self.SLEEP_TIME)
            current_wait_time += self.SLEEP_TIME
        # not raise error in case of timeout parameter
        if timeout:
            return False
        else:
            raise ElasticSearchDataNotUploaded

    @typechecked()
    def upload_to_elasticsearch(self, index: str, data: dict, doc_type: str = '_doc', es_add_items: dict = None, **kwargs):
        """
        This method uploads json data into elasticsearch
        :param index: index name to be stored in elasticsearch
        :param data: data must be in dictionary i.e. {'key': 'value'}
        :param doc_type:
        :param es_add_items:
        :return:
        """
        if index and data:
            # Add items
            if es_add_items:
                data.update(es_add_items)

            # utcnow - solve timestamp issue
            data['timestamp'] = datetime.now(timezone.utc)  # datetime.utcnow() or datetime.now()

            # Uploads data to elasticsearch server
            try:
                if isinstance(data, dict):  # JSON Object
                    self.__es.index(index=index, doc_type=doc_type, document=data, **kwargs)
                else:  # JSON Array
                    for record in data:
                        self.__es.index(index=index, doc_type=doc_type, document=record, **kwargs)
                return True
            except Exception as err:
                raise err
        else:
            raise Exception('Empty parameters: index/ data')

    @typechecked()
    @logger_time_stamp
    def update_elasticsearch_index(self, index: str, id: str, metadata: dict = ''):
        """
        This method updates existing index with specific id
        :param index: index name
        :param id: The specific id
        :param metadata: The metadata for enrich that existing index according to id
        :return:
        """
        if index and id and metadata:
            self.__es.update(index=index, id=id, body={"doc": metadata})
        else:
            raise Exception('Empty parameters: index/ id/ metadata')

    @typechecked()
    @logger_time_stamp
    def get_elasticsearch_index_by_id(self, index: str, id: str):
        """
        This method returns elasticsearch index data by id
        :param index: index name
        :param id: The specific index id
        :return:
        """
        if index and id:
            return self.__es.get(index=index, id=id)
        else:
            raise Exception('Empty parameters: index/ id')

    @typechecked()
    @logger_time_stamp
    def delete_elasticsearch_index_by_id(self, index: str, id: str):
        """
        This method deletes elasticsearch index data by id
        :param index: index name
        :param id: The specific index id
        :return:
        """
        if index and id:
            return self.__es.delete(index=index, id=id)
        else:
            raise Exception('Empty parameters: index/ id')

    @typechecked()
    @logger_time_stamp
    def get_query_data_between_dates(self, start_datetime: datetime, end_datetime: datetime):
        """
        This method returns the query for fetching data between dates
        @param start_datetime:
        @param end_datetime:
        @return:
        """
        if start_datetime and end_datetime:
            if end_datetime < start_datetime:
                start_datetime = end_datetime
            query = {
                    "bool": {
                        "filter": {
                            "range": {
                                "timestamp": {
                                    "format": "yyyy-MM-dd HH:mm:ss"
                                }
                            }
                        }
                    }
            }
            query['bool']['filter']['range']['timestamp']['lte'] = str(end_datetime.replace(microsecond=0))
            query['bool']['filter']['range']['timestamp']['gte'] = str(start_datetime.replace(microsecond=0))
            return query
        else:
            raise Exception('Empty parameters: start_datetime/ end_datetime')

    @typechecked()
    @logger_time_stamp
    def get_index_data_between_dates(self, index: str, start_datetime: datetime = None, end_datetime: datetime = None):
        """
        This method returns list of index data between dates
        @param index:
        @param start_datetime:
        @param end_datetime:
        @return: list of ids
        """
        number_of_documents = 100
        scroll_duration = '1h'
        if index and start_datetime and end_datetime:
            es_data = []
            query = self.get_query_data_between_dates(start_datetime, end_datetime)
            response = self.__es.search(index=index, query=query, size=number_of_documents, scroll=scroll_duration)
            scroll_id = response.get('_scroll_id')
            if response.get('hits').get('hits'):
                es_data.extend(response.get('hits').get('hits'))
            while scroll_id:
                response = self.__es.scroll(scroll_id=scroll_id, scroll="1h")
                if len(response.get('hits').get('hits')) > 0:
                    es_data.extend(response.get('hits').get('hits'))
                else:
                    break
            return es_data
        else:
            raise Exception('Empty parameters: index/ start_datetime/ end_datetime')

    @typechecked()
    @logger_time_stamp
    def get_index_ids_between_dates(self, index: str, start_datetime: datetime = None, end_datetime: datetime = None):
        """
        This method returns list of index ids between dates
        @param index:
        @param start_datetime:
        @param end_datetime:
        @return: list of ids
        """
        es_data = self.get_index_data_between_dates(index=index, start_datetime=start_datetime, end_datetime=end_datetime)
        ids = [hit["_id"] for hit in es_data]
        return ids

    @typechecked()
    @logger_time_stamp
    def delete_index_ids_between_dates(self, index: str, start_datetime: datetime = None,
                                       end_datetime: datetime = None):
        """
        This method deletes index ids between dates
        @param index:
        @param start_datetime:
        @param end_datetime:
        @return:
        """
        if index and start_datetime and end_datetime:
            ids = self.get_index_ids_between_dates(index=index, start_datetime=start_datetime, end_datetime=end_datetime)
            for id in ids:
                self.delete_elasticsearch_index_by_id(index=index, id=id)
        else:
            raise Exception('Empty parameters: index/ start_datetime/ end_datetime')

    @logger_time_stamp
    def get_all_indexes(self):
        """
        This method returns sorted elasticsearch indexes
        @return:
        """
        indexes = []
        # Fetch all indices
        indices = self.__es.cat.indices(format="json")
        for index in indices:
            indexes.append(index['index'])
        return sorted(indexes)




import time
import os
import re
from typeguard import typechecked
from benchmark_runner.common.prometheus.prometheus_snapshot_exceptions import PrometheusSnapshotAlreadyStarted, PrometheusSnapshotNotStarted, PrometheusSnapshotAlreadyRetrieved
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.oc.oc import OC


class PrometheusSnapshot:
    """
    Extract a Prometheus snapshot into a specified directory.
    The directory is assumed to already exist
    """

    # Constants
    # Default snapshot date/time format, similar to ISO
    prometheus_snapshot_datetime_format = '%Y_%m_%dT%H_%M_%S%z'
    # Pattern to recognize valid date/time.  If we try to get a timestamp before
    # the prometheus pod is fully initialized, we'll get an error back; this is
    # to distinguish a valid date from an error.  If the datetime format is changed,
    # this pattern may have to be changed too.
    date_fast_recognition_pattern = r'^[0-9]{4}'
    # How many times to retry getting a valid date/time before we give up
    timestamp_retries = 12
    # How long to sleep between each attempt to get the date/time
    timestamp_sleep = 5
    # Default time to wait after the prometheus pod comes up before
    # we return to the caller, to allow collection of baseline
    # metrics
    default_prewait_time = 60
    # Default time to wait after we receive the request to provide a snapshot, to allow
    # metrics data to fully settle down
    default_postwait_time = 120

    def __init__(self, oc: OC, artifacts_path: str, verbose: bool=False):
        self.__oc = oc
        self.__artifacts_path = artifacts_path
        self.__start_timestamp = None
        self.__end_timestamp = None
        self.__verbose = verbose
        self.__pod_name = 'prometheus-k8s-0'
        self.__namespace = 'openshift-monitoring'
        self.__container = 'prometheus'
        self.__format = self.prometheus_snapshot_datetime_format

    @typechecked
    def __verbose_log(self, message: str):
        """
        Log message if verbosity is enabled
        :param message:
        """
        if self.__verbose:
            logger.info(message)

    @typechecked
    def __exec(self, command: str):
        """
        Run a command on the Prometheus pod (by default)
        :param command:
        """
        return self.__oc.exec(pod_name=self.__pod_name, namespace=self.__namespace, container=self.__container, command=command)

    def __prom_timestamp(self):
        """
        Retrieve a timestamp from the prometheus pod.
        We use a timestamp from the pod because timestamps within the metrics db
        will be based on the time (including timezone) within the pod, not that on the host.
        """
        valid_timestamp = False
        pattern = self.date_fast_recognition_pattern
        retries = self.timestamp_retries
        # The prometheus pod may show as ready before the prometheus container actually
        # is ready (ready does not appear to indicate that all containers are necessarily
        # started).  So retry for up to a minute before giving up.
        while not valid_timestamp:
            timestamp = self.__exec(f"date '+{self.prometheus_snapshot_datetime_format}'")
            if re.match(pattern, timestamp):
                return timestamp
            else:
                time.sleep(self.timestamp_sleep)
                if retries <= 0:
                    return timestamp
                retries = retries - 1

    @typechecked
    @logger_time_stamp
    def prepare_for_snapshot(self, pre_wait_time: int=default_prewait_time, clear_db: bool=True):
        """
        Prepare for a prometheus database snapshot.  Must be called prior to running
        the workload.
        :param pre_wait_time:
        :param clear_db:
        """
        if self.__start_timestamp or self.__end_timestamp:
            raise PrometheusSnapshotAlreadyStarted
        if clear_db:
            self.__verbose_log(f'Deleting existing {self.__pod_name} (Prometheus snapshot) pod')
            self.__oc.terminate_pod_sync(pod_name=self.__pod_name, namespace=self.__namespace)
        self.__verbose_log(f'Waiting for {self.__pod_name} pod to reappear')
        self.__oc.wait_for_pod_ready(pod_name=self.__pod_name, namespace=self.__namespace)
        self.__start_timestamp = self.__prom_timestamp()
        self.__verbose_log(f'Waiting {pre_wait_time} seconds for Prometheus snapshot to fully initialize')
        time.sleep(pre_wait_time)

    @typechecked
    def retrieve_snapshot(self, post_wait_time: int=default_postwait_time):
        """
        Retrieve the prometheus database snapshot after running the workload.
        The filename that the snapshot is placed in is returned.
        :param post_wait_time:
        """
        if not self.__start_timestamp:
            raise PrometheusSnapshotNotStarted
        if self.__end_timestamp:
            raise PrometheusSnapshotAlreadyRetrieved
        self.__verbose_log(f'Waiting {post_wait_time} seconds for Prometheus snapshot to complete')
        time.sleep(post_wait_time)
        self.__end_timestamp = self.__prom_timestamp()
        promdb_name = f'promdb_{self.__start_timestamp}_{self.__end_timestamp}'
        promdb_path = os.path.join(self.__artifacts_path, f'{promdb_name}.tar')
        xformcmd = f"--transform 's,^[.],./{promdb_name},'"
        self.__verbose_log(f'Saving prometheus snapshot to {promdb_path}')
        cmd = f'/bin/sh -c "tar -C /prometheus {xformcmd} -cf - .; true" > "{promdb_path}"'
        self.__exec(cmd)
        return promdb_path




class IBMError(Exception):
    """ Base class for all IBM error classes.
        All exceptions raised by the benchmark runner library should inherit from this class. """
    pass


class IBMMachineNotLoad(IBMError):
    """
    This class exception is for Subprocess
    """
    def __init__(self):
        self.message = 'The IBM machine did not load'
        super(IBMMachineNotLoad, self).__init__(self.message)




import time
from typeguard import typechecked

from benchmark_runner.common.oc.oc import OC
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp
from benchmark_runner.main.environment_variables import environment_variables
from benchmark_runner.common.ocp_resources.create_ocp_resource_exceptions import OCPResourceCreationTimeout


class CreateOCPResourceOperations:
    """
    This class creates OCP resources
    """
    # Expected CSVs List names
    EXPECTED_ODF_CSV = ['mcg-operator', 'ocs-operator', 'odf-csi-addons-operator', 'odf-operator']
    SLEEP_TIME = 30

    def __init__(self, oc: OC):
        self._environment_variables_dict = environment_variables.environment_variables_dict
        self.__oc = oc

    @staticmethod
    @typechecked
    def _replace_in_file(file_path: str, old_value: str, new_value: str):
        """
        This method replaces string in file
        :param file_path:
        :param old_value:
        :param new_value:
        :return:
        """
        # Read in the file
        with open(file_path, 'r') as file:
            file_data = file.read()

        # Replace the target string
        file_data = file_data.replace(old_value, new_value)

        # Write the file out again
        with open(file_path, 'w') as file:
            file.write(file_data)

    @typechecked
    @logger_time_stamp
    def wait_for_ocp_resource_create(self, operator: str, verify_cmd: str, status: str = '', count_disk_maker: bool = False, count_openshift_storage: bool = False, kata_worker_machine_count: bool = False, count_csv_names: bool = False, timeout: int = int(environment_variables.environment_variables_dict['timeout'])):
        """
        This method waits till operator is created or throw exception after timeout
        :param operator: The operator cnv, lso, odf, kata
        :param verify_cmd: Verify command that resource was created successfully
        :param status: The final success status
        :param count_disk_maker: count disk maker
        :param count_openshift_storage: count openshift storage disks
        :param kata_worker_machine_count: count kata worker machine
        :param timeout: Timeout duration for OpenShift resource creation.
        :return: True if met the result
        """
        current_wait_time = 0
        double_check = False
        while timeout <= 0 or current_wait_time <= timeout:
            cmd = self.__oc.run(verify_cmd)
            # Count openshift-storage/ pv
            if count_openshift_storage:
                if int(cmd) == self.__oc.get_num_active_nodes() * int(environment_variables.environment_variables_dict['num_odf_disk']):
                    return True
            # Count disk maker (worker/master number * disk maker)
            elif count_disk_maker:
                if int(cmd) == int(self.__oc.get_num_active_nodes()) * 2:
                    return True
            # Count worker machines
            elif kata_worker_machine_count:
                if int(cmd) > 0:
                    return True
            # Count CSV
            elif count_csv_names:
                # Wait to CSV
                if cmd:
                    # ODF operator: not all CSV are started on the same time
                    if operator == 'odf':
                        # Verify each CSV name
                        if all(any(actual_csv.startswith(expected_csv) for actual_csv in cmd.split()) for expected_csv in self.EXPECTED_ODF_CSV):
                            return True
                    # default 1 CSV
                    else:
                        return True
            elif status:
                # catch equal or contains
                if status in cmd:
                    if double_check:
                        return True
                    else:
                        double_check = True
                        # sleep for x seconds
                        time.sleep(self.SLEEP_TIME)
                        current_wait_time += self.SLEEP_TIME
                        continue
                else:
                    double_check = False
            # sleep for x seconds
            time.sleep(self.SLEEP_TIME)
            current_wait_time += self.SLEEP_TIME
        raise OCPResourceCreationTimeout(operator)

    def verify_csv_installation(self, namespace: str, operator: str, upgrade_version: str = None, csv_num: int = 1):
        """
        This method verifies csv installation
        :param namespace:
        :param operator:
        :param csv_num: number of csvs, default 1
        :param upgrade_version: in upgrade mode, check for upgrade version
        """
        if upgrade_version:
            self.__oc.wait_for_upgrade_version(operator=operator, upgrade_version=upgrade_version, namespace=namespace)
        csv_names = self.__oc.wait_for_csv(operator=operator, csv_num=csv_num, namespace=namespace)
        for csv_name in csv_names.split():
            self.wait_for_ocp_resource_create(operator=operator,
                                              verify_cmd=f"oc get csv -n {namespace} {csv_name} -o jsonpath='{{.status.phase}}'",
                                              status='Succeeded')

    def apply_patch(self, namespace: str, operator: str):
        """
        This method applies an InstallPlan that has not been approved
        :param namespace:
        :param operator:
        :return:
        """
        install_plan_cmd = f"oc get InstallPlan -n {namespace} -o jsonpath={{$.items[*].metadata.name}}"
        install_plan_names = self.__oc.run(cmd=install_plan_cmd).split()
        for name in install_plan_names:
            check_approved = f"oc get InstallPlan -n {namespace} {name} -o jsonpath={{..spec.approved}}"
            approved = self.__oc.run(cmd=check_approved)
            # APPROVED false - need to patch
            if not {'true': True, 'false': False}.get(approved.lower(), False):
                install_plan_cmd = (f"oc patch InstallPlan -n {namespace} {name} -p '{{\"spec\":{{\"approved\":true}}}}' --type merge")
                self.__oc.run(cmd=install_plan_cmd)
                # Wait till installPlan is approved
                self.wait_for_ocp_resource_create(operator=operator,
                                                  verify_cmd=check_approved,
                                                  status="true")
                # Wait till CSV name is created
                self.wait_for_ocp_resource_create(operator=operator,
                                                  verify_cmd=f"oc get csv -n {namespace} -o jsonpath={{$.items[*].metadata.name}}",
                                                  count_csv_names=True)
                # Verify CSV installation
                self.verify_csv_installation(namespace=namespace, operator=operator)




import os
import sys
import time
from multiprocessing import Process

from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.elasticsearch.elasticsearch_exceptions import ElasticSearchDataNotUploaded
from benchmark_runner.workloads.bootstorm_vm import BootstormVM


class WindowsVM(BootstormVM):
    """
    This class runs Windows vm
    """
    def __init__(self):
        super().__init__()
        if not self._windows_url:
            raise ValueError('Missing Windows DV URL')

    @logger_time_stamp
    def run(self):
        """
        This method runs the workload
        :return:
        """
        try:
            if self._run_type == 'test_ci':
                self._es_index = 'windows-test-ci-results'
            else:
                self._es_index = 'windows-results'
            self._initialize_run()
            if not self._verification_only:
                # create windows dv
                self._oc.create_async(yaml=os.path.join(f'{self._run_artifacts_path}', 'windows_dv.yaml'))
                self._oc.wait_for_dv_status(status='Succeeded')
            self.run_vm_workload()
            if self._delete_all:
                # delete windows dv
                self._oc.delete_async(yaml=os.path.join(f'{self._run_artifacts_path}', 'windows_dv.yaml'))
                # delete namespace
                self._oc.delete_async(yaml=os.path.join(f'{self._run_artifacts_path}', 'namespace.yaml'))
        except ElasticSearchDataNotUploaded as err:
            self._oc.delete_vm_sync(
                yaml=os.path.join(f'{self._run_artifacts_path}', f'{self._name}.yaml'),
                vm_name=self._vm_name)
            raise err
        except Exception as err:
            # save run artifacts logs
            if self._es_host:
                self._data_dict['run_artifacts_url'] = os.path.join(self._run_artifacts_url, f'{self._get_run_artifacts_hierarchy(workload_name=self._workload_name, is_file=True)}-{self._time_stamp_format}.tar.gz')
                self._upload_to_elasticsearch(index=self._es_index, kind=self._kind, status='failed', result=self._data_dict)
                # verify that data upload to elastic search according to unique uuid
                self._verify_elasticsearch_data_uploaded(index=self._es_index, uuid=self._uuid)
            raise err




import os
import json

from benchmark_runner.common.ssh.ssh import SSH
from benchmark_runner.clusterbuster.clusterbuster_exceptions import MissingResultReport, MissingElasticSearch
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.workloads.workloads_operations import WorkloadsOperations


class ClusterBusterWorkloads(WorkloadsOperations):
    """
    This class is responsible for all ClusterBuster workloads
    """

    def __init__(self):
        super().__init__()
        self.__clusterbuster_path = '/tmp/OpenShift4-tools/CI/./run-kata-perf-suite'
        # environment variables
        self.__namespace = self._environment_variables_dict.get('namespace', '')
        self.__clusterbuster_workload = self._environment_variables_dict.get('clusterbuster_workload', '')
        self.__clusterbuster_uuid = self._environment_variables_dict.get('clusterbuster_uuid', '')
        self.__result_report = '/tmp/clusterbuster-report.json'
        self.__artifactdir = os.path.join(self._run_artifacts_path, 'clusterbuster-ci')
        self.__clusterbuster_log = os.path.join(self._run_artifacts_path, 'clusterbuster.log')
        self.__ssh = SSH()

    @logger_time_stamp
    def upload_clusterbuster_result_to_elasticsearch(self):
        """
        This method upload to ElasticSearch the results
        :return:
        """
        result_report_json_file = open(self.__result_report)
        result_report_json_str = result_report_json_file.read()
        result_report_json_data = json.loads(result_report_json_str)
        for workload, clusterbuster_tests in result_report_json_data.items():
            if self._run_type == 'test_ci':
                index = f'clusterbuster-{workload}-test-ci-results'
            elif self._run_type == 'release':
                index = f'clusterbuster-{workload}-release-results'
            else:
                index = f'clusterbuster-{workload}-results'
            logger.info(f'upload index: {index}')
            if workload != 'metadata':
                for clusterbuster_test in clusterbuster_tests:
                    self._es_operations.upload_to_elasticsearch(index=index, data=clusterbuster_test)
            # metadata
            elif workload == 'metadata':
                # run artifacts data
                result_report_json_data['metadata']['run_artifacts_url'] = os.path.join(self._run_artifacts_url, f'{self._get_run_artifacts_hierarchy(workload_name=self._workload, is_file=True)}-{self._time_stamp_format}.tar.gz')
                self._es_operations.upload_to_elasticsearch(index=index, data=result_report_json_data['metadata'])
                self._es_operations.verify_elasticsearch_data_uploaded(index=index, uuid=result_report_json_data['metadata']['uuid'])

    @logger_time_stamp
    def delete_all(self):
        """
        This method delete all resource that related to ClusterBuster resource
        :return:
        """
        self.__ssh.run(cmd=f'oc delete pod -A -l {self.__namespace}')

    def initialize_workload(self):
        """
        This method includes all the initialization of ClusterBuster workload
        :return:
        """
        self.delete_all()
        self.clear_nodes_cache()
        if self._enable_prometheus_snapshot:
            self.start_prometheus()

    def finalize_workload(self):
        """
        This method includes all the finalization of ClusterBuster workload
        :return:
        """
        # Result file exist and not empty
        if os.path.exists(os.path.join(self.__result_report)) and not os.stat(self.__result_report).st_size == 0:
            self.upload_clusterbuster_result_to_elasticsearch()

        if self._enable_prometheus_snapshot:
            self.end_prometheus()
        if self._endpoint_url:
            self.upload_run_artifacts_to_s3()
        if not self._save_artifacts_local:
            self.delete_local_artifacts()
        self.delete_all()

    @logger_time_stamp
    def run_workload(self):
        """
        This method run ClusterBuster workload
        :return:
        """
        # release pass uuid and workload
        if self._run_type == 'release':
            self.__ssh.run(cmd=f'{self.__clusterbuster_path} --run_type={self._run_type} --client-pin-node={self._pin_node1} --server-pin-node={self._pin_node2} --sync-pin-node={self._pin_node2} --force-cleanup-i-know-this-is-dangerous=600 --basename={self.__namespace} --artifactdir={self.__artifactdir} --analyze={self.__result_report} --uuid={self.__clusterbuster_uuid} {self.__clusterbuster_workload} > {self.__clusterbuster_log} 2>&1')
        else:
            self.__ssh.run(cmd=f'{self.__clusterbuster_path} --run_type={self._run_type} --client-pin-node={self._pin_node1} --server-pin-node={self._pin_node2} --sync-pin-node={self._pin_node2} --force-cleanup-i-know-this-is-dangerous=600 --basename={self.__namespace} --artifactdir={self.__artifactdir} --analyze={self.__result_report} > {self.__clusterbuster_log} 2>&1')
        # Check that Result file exist and not empty
        if os.path.exists(os.path.join(self.__result_report)) and not os.stat(self.__result_report).st_size == 0:
            self.__ssh.run(cmd=f'cp {self.__result_report} {self._run_artifacts_path}')
            return True
        else:
            result_report_json_data = {}
            result_report_json_data['result'] = 'Failed'
            result_report_json_data['run_artifacts_url'] = os.path.join(self._run_artifacts_url, f'{self._get_run_artifacts_hierarchy(workload_name=self._workload, is_file=True)}-{self._time_stamp_format}.tar.gz')
            if self._run_type == 'test_ci':
                index = f'clusterbuster-metadata-test-ci-results'
            elif self._run_type == 'release':
                index = f'clusterbuster-metadata-release-results'
            else:
                index = f'clusterbuster-metadata-results'
            logger.info(f'upload index: {index}')
            self._es_operations.upload_to_elasticsearch(index=index, data=result_report_json_data)
            raise MissingResultReport()

    @logger_time_stamp
    def run(self):
        """
        This method run ClusterBuster workloads
        :return:
        """
        try:
            # initialize workload
            self.initialize_workload()
            # Run workload
            if self.run_workload():
                # finalize workload
                self.finalize_workload()
        # when error raised finalize workload
        except Exception:
            logger.info(f'{self._workload} workload raised an exception')
            # finalize workload
            self.finalize_workload()
            return False

        return True




class OCPResourceError(Exception):
    """ Base class for all OCP resource error classes. """
    pass


class OCPResourceCreationTimeout(OCPResourceError):
    """This exception returns resource creation timeout error"""
    def __init__(self, resource):
        self.message = f'The {resource} resource was not created'
        super(OCPResourceCreationTimeout, self).__init__(self.message)


class KataInstallationFailed(OCPResourceError):
    """This exception returns failure to install sandboxed containers"""
    def __init__(self, reason):
        self.message = f'Installation of sandboxed containers failed: {reason}'
        super(KataInstallationFailed, self).__init__(self.message)


class ODFInstallationFailed(OCPResourceError):
    """This exception returns failure to install ODF"""
    def __init__(self):
        self.message = f'ODF installation failed'
        super(ODFInstallationFailed, self).__init__(self.message)




import os

from benchmark_runner.common.oc.oc import OC
from benchmark_runner.common.logger.logger_time_stamp import logger_time_stamp, logger
from benchmark_runner.common.ocp_resources.create_ocp_resource_operations import CreateOCPResourceOperations


class CreateCNV(CreateOCPResourceOperations):
    """
    This class is created CNV nightly build operator
    """
    def __init__(self, oc: OC, path: str, resource_list: list):
        super().__init__(oc)
        self.__oc = oc
        self.__path = path
        self.__resource_list = resource_list

    @logger_time_stamp
    def create_cnv(self):
        """
        This method creates cnv resource
        :return:
        """
        cnv_version = self._environment_variables_dict.get('cnv_version', '')
        cnv_nightly_channel = self._environment_variables_dict.get('cnv_nightly_channel', '')
        if cnv_nightly_channel:
            logger.info('wait for cnv-nightly kubevirt-hyperconverged')
            # cnv_nightly_catalog_source.yaml already run direct through Github actions - just need to wait
            self.wait_for_ocp_resource_create(operator='cnv_nightly_catalog_source.yaml',
                                              verify_cmd="oc get packagemanifest -l catalog=cnv-nightly-catalog-source | grep kubevirt-hyperconverged",
                                              status="kubevirt-hyperconverged")
            starting_csv = self.__oc.run(""" oc get packagemanifest -l "catalog=cnv-nightly-catalog-source" -o jsonpath="{$.items[?(@.metadata.name=='kubevirt-hyperconverged')].status.channels[?(@.name==\\"nightly-cnv_version\\")].currentCSV}" """.replace('cnv_version', cnv_version))
            self._replace_in_file(file_path=os.path.join(self.__path, '01_operator.yaml'), old_value="@starting_csv@", new_value=starting_csv)
        for resource in self.__resource_list:
            logger.info(f'run {resource}')
            if resource.endswith('.sh'):
                self.__oc.run(cmd=f'chmod +x {os.path.join(self.__path, resource)}; {self.__path}/./{resource}')
            else:
                self.__oc.create_async(yaml=os.path.join(self.__path, resource))
                if '01_operator.yaml' in resource:
                    if cnv_nightly_channel:
                        # wait till get the patch
                        self.wait_for_ocp_resource_create(operator=resource,
                                                          verify_cmd="oc get InstallPlan -n openshift-cnv -o jsonpath={.items[0].metadata.name}",
                                                          status="install-")
                        self.apply_patch(namespace='openshift-cnv', operator='cnv')
                    # stable channel
                    else:
                        self.wait_for_ocp_resource_create(operator='cnv',
                                                          verify_cmd="oc get csv -n openshift-cnv -o jsonpath='{.items[0].status.phase}'",
                                                          status='Succeeded')
                # for second script wait for refresh status
                if '02_hyperconverge.yaml' in resource:
                    # Wait that till succeeded
                    self.wait_for_ocp_resource_create(operator='cnv',
                                                      verify_cmd="oc get csv -n openshift-cnv -o jsonpath='{.items[0].status.phase}'",
                                                      status='Succeeded')
        return True
